import logging
from functools import partial
from pathlib import Path
from typing import Any, Callable, Optional, Sequence, Type, Union

import numpy as np
import pandas as pd

from .. import utils
from .base import BaseAnnotator, BaseAnnotatorJSON, SingleAnnotator

__all__ = ["PairwiseAnnotator", "SinglePairwiseAnnotator"]

PAIRWISE_ADDED_DOCSTRING = """
    
    p_label_flip : float, optional
        Probability of flipping the label (ie adds noise by taking a mixture between predicted label and
        2*p_label_flip of independent coin flip). If None, will not flip the label. In AlpacaFarm we use 0.25
        for training. You can set this later on using `set_noise`.
        
    input_keys : sequence of str, optional
        Keys use to distinguish inputs.

    output_keys : sequence of str, optional
        Keys use to distinguish outputs.
        
    Notes
    -----
    There are three main functions for annotations depending on how the outputs to compare are given:
        - annotate_pairs: annotate a sequence of examples that contain the pair of outputs `"output_1"` and `"output_2"`
        - annotate_samples: annotate a sequence of examples that contain `"output"` from which we will sample a pair of
            outputs. Useful for collecting pairwise preferences for RLHF.
        - annotate_head2head: annotate a pair of sequence of outputs, each containing `"output"` which will be merged
            into a single sequence of paired outputs. Useful for evaluation against a reference.
    """


class PairwiseAnnotatorLocal(BaseAnnotator):
    __doc__ = BaseAnnotator.__doc__.replace("Base class", "Class") + PAIRWISE_ADDED_DOCSTRING

    def __init__(
        self,
        *args,
        input_keys: Sequence[str] = ("instruction",),
        output_keys: Sequence[str] = ("output_1", "output_2"),
        p_label_flip: Optional[float] = None,
        **kwargs,
    ):
        self.input_keys = list(input_keys)
        self.output_keys = list(output_keys)
        super().__init__(*args, **kwargs, primary_keys=self.input_keys + self.output_keys)
        self.p_label_flip = p_label_flip

    @property
    def SingleAnnotator(self) -> Type["SingleAnnotator"]:
        return SinglePairwiseAnnotator

    @property
    def annotation_key(self) -> str:
        return "preference"

    @property
    def random_seed_keys(self) -> list[str]:
        return list(self.input_keys)

    def annotate_samples(
        self,
        all_outputs: utils.AnyData,
        keys_to_sample_output_2: Optional[Sequence] = None,
        is_unique_instructions: bool = True,
        p_label_flip: Optional[float] = None,
        is_multisample_list: bool = True,
        **decoding_kwargs,
    ) -> list[dict[str, Any]]:
        """Sample pairs of outputs from a sequence of examples and annotate them.

        Parameters
        ----------
        all_outputs : list of dict or pd.DataFrame or datasets.Dataset
            All examples from which we will sample a pair of outputs to annotate. Each dictionary (or row) should
            contain all of `self.input_keys` and `keys_to_sample_output_2` and `"output"`.

        keys_to_sample_output_2 : tuple of str, optional
            Keys to use to sample paired `"output_2"` to compare to the current `"output"` which will become
            `"output_1"`. If `None` it uses `self.input_keys`.

        is_unique_instructions : bool, optional
            Whether to deduplicate the instructions such that there is only one pair per instruction. If False
            there will be as many pairs as there are outputs for each instruction.

        p_label_flip : float, optional
            Probability of flipping the label (ie adds noise by taking a mixture between predicted label and
            2*p_label_flip of independent coin flip). If None, will use `self.p_label_flip`.

        is_multisample_list : bool, optional
            If True `all_outputs` is a list of examples (dictionary) and each example has an `"output"` column
            containing
            a list of all multi samples. If False `"output"` contains a single output but each element in the list is a
            different (instruction, output) pair with potentially the same instruction.

        decoding_kwargs :
            Additional arguments to pass to the decoder.
        """

        all_outputs = utils.convert_to_dataframe(all_outputs)

        if is_multisample_list:
            all_outputs = all_outputs.explode("output").reset_index().rename(columns={"index": "sample_id"})
            all_outputs["sample_id"] = all_outputs.groupby("sample_id").cumcount()

        if keys_to_sample_output_2 is None:
            keys_to_sample_output_2 = self.input_keys
        keys_to_sample_output_2 = list(keys_to_sample_output_2)

        n_pre_drop = len(all_outputs)

        # set output to be unique for each keys_to_sample_output_2
        df_to_annotate = (
            all_outputs.groupby(keys_to_sample_output_2)
            .apply(lambda x: x.drop_duplicates(["output"]))
            .reset_index(drop=True)
            .rename(columns={"output": "output_1"})
        )

        if len(df_to_annotate) != n_pre_drop:
            logging.warning(
                f"Filtered rows because of duplicate outputs for the same keys_to_sample_output_2="
                f"{keys_to_sample_output_2}. {n_pre_drop} -> {len(df_to_annotate)}"
            )

        # sample an output 2 for each output 1 that are different
        df_to_annotate["output_2"] = df_to_annotate.groupby(list(keys_to_sample_output_2))["output_1"].transform(
            lambda x: utils.random_derangement(x.values, seed=self.seed)
        )

        if is_unique_instructions:
            n_pre_dedup = len(df_to_annotate)
            df_to_annotate = df_to_annotate.drop_duplicates(subset=self.input_keys)
            if len(df_to_annotate) != n_pre_dedup:
                logging.info(f"Filtered unique instruction/input pairs: {n_pre_dedup} -> {len(df_to_annotate)}")

        if p_label_flip is not None:
            old_p_label_flip = self.p_label_flip
            self.set_noise(p_label_flip)

        try:
            annotated = self.__call__(df_to_annotate, **decoding_kwargs)
        finally:
            # reset even if there is an error
            if p_label_flip is not None:
                self.set_noise(old_p_label_flip)

        return annotated

    def annotate_head2head(
        self,
        outputs_1: Union[Sequence[dict[str, Any]], pd.DataFrame],
        outputs_2: Union[Sequence[dict[str, Any]], pd.DataFrame],
        keys_to_merge: Optional[Sequence[str]] = None,
        is_ordered: bool = False,
        **decoding_kwargs,
    ) -> list[dict[str, Any]]:
        """Head-to-head comparison between two sequence of outputs.

        Parameters
        ----------
        outputs_1 : list of dict or dataframe
            Examples to annotate. Each dictionary (or row) should contain all of `keys_to_merge` and `"output"`.
            `"output"` will become `"output_1"`.

        outputs_2 : list of dict or dataframe
            Second  to annotate. Each dictionary (or row) should contain all of `keys_to_merge` and `"output"`.
            `"output"` will become `"output_2"`.

        keys_to_merge : tuple of str, optional
            Keys to use to merge the two sequences of outputs. If None uses `self.input_keys`

        is_ordered : bool, optional
            Whether the two sequences of outputs are in matching order. If not we will be merging based on
            `keys_to_merge`, which means that the outputs can actually be shorter than the inputs (if some outputs
            are not found in the other sequence) or longer (if some outputs are duplicated in both sequences =>
            set cartesian products).

        decoding_kwargs :
            Additional arguments to pass to `fn_completions`.

        Returns
        -------
        annotated : list of dict
            The annotated examples. Each dictionary will contain all of `keys_to_merge`, `"output_1"`, `"output_2"`, and
            `"preference"`. Preference will be 1.5 if output_1 == output_2, 1 if output_1 is preferred, and 2 if output_2
            is preferred.
        """
        if keys_to_merge is None:
            keys_to_merge = self.input_keys

        keys_to_merge = list(keys_to_merge)

        outputs_1 = utils.convert_to_dataframe(outputs_1)
        outputs_2 = utils.convert_to_dataframe(outputs_2)

        if is_ordered:
            outputs_1 = outputs_1.copy()
            outputs_2 = outputs_2.copy()
            outputs_1["tmp_idx"] = range(len(outputs_1))
            outputs_2["tmp_idx"] = range(len(outputs_1))
            keys_to_merge += ["tmp_idx"]  # add a temporary index to merge on

        # find all the columns that are in both
        other_same_cols = [k for k in outputs_1.columns if k in outputs_2 and k not in (keys_to_merge + ["output"])]

        df_to_annotate = pd.merge(
            outputs_1,
            outputs_2,
            on=keys_to_merge,
            suffixes=("_1", "_2"),
        )

        for c in other_same_cols:
            # if the columns are the same, we can drop the _2
            if df_to_annotate[c + "_1"].equals(df_to_annotate[c + "_2"]):
                df_to_annotate = df_to_annotate.drop(columns=c + "_2").rename(columns={c + "_1": c})

        if is_ordered:
            df_to_annotate = df_to_annotate.drop(columns="tmp_idx")
        else:
            # if you are taking the cartesian product, you can have undesired duplicates
            df_to_annotate = df_to_annotate.drop_duplicates()

            if not (len(outputs_1) == len(outputs_2) == len(df_to_annotate)):
                logging.warning(
                    f"The length of outputs before and after merge are not the same. We have len(outputs_1)=="
                    f"{len(outputs_1)}, len(outputs_2)=={len(outputs_2)}, and len(df_annotated)=={len(df_to_annotate)}."
                    f" This means that there are missing examples or duplicates. We are taking a SQL inner join."
                )

        out = self.__call__(df_to_annotate, **decoding_kwargs)

        return out

    def annotate_pairs(
        self,
        to_annotate: Union[Sequence[dict[str, Any]], pd.DataFrame],
        **decoding_kwargs,
    ) -> list[dict[str, Any]]:
        """Annotates the given examples, which contain both `"output_1"` and `"output_2"` keys.

        Parameters
        ----------
        to_annotate : list of dict or dataframe
            Examples to annotate. Each dictionary (or row) should contain all of `self.primary_keys`.

        **decoding_kwargs :
            Additional arguments to pass to `fn_completions`.

        Returns
        -------
        annotated : list of dict
            The annotated examples. Each dictionary will contain all of `self.primary_keys` and `"preference"`.
            Preference will be 0 if output_1 == output_2, 1 if output_1 is preferred, and 2 if output_2 is preferred.
        """
        # `annotate_pairs` is used for backward compatibility
        return self.__call__(to_annotate, **decoding_kwargs)

    def set_noise(self, p_label_flip: float):
        """Set the noise level for the annotators.

        Parameters
        ----------
        p_label_flip : float, optional
            Probability of flipping the label (ie adds noise by taking a mixture between predicted label and
            2*p_label_flip of independent coin flip). If None, will not flip the label. In AlpacaFarm we use 0.25
            for training.
        """
        self.p_label_flip = p_label_flip

    def _preprocess(self, to_annotate: utils.AnyData) -> pd.DataFrame:
        # same as preprocess but with potential random noising and dealing with eauality

        df_to_annotate = super()._preprocess(to_annotate)

        # 1. adds random noise => avoids annotating examples that will be noised out.
        if self.p_label_flip:
            logging.info(f"Adding random noise to the labels p_label_flip={self.p_label_flip}.")
            # if you have 25% change of flipping the label, you have 50% chance of selecting random label
            # note that the noise is always binary (1 or 2), even when the annotation is a float (e.g. using logprobs)
            p_noise = self.p_label_flip * 2
            noisy_preference = df_to_annotate.apply(
                # we add "noisy_label" at the beginning to use ~independent seeds between tasks
                lambda x: utils.random_seeded_choice(  # seed on inputs for reproducibility
                    seed="noisy_preference" + "".join(x[self.random_seed_keys]) + str(self.seed),
                    choices=[np.nan, 1, 2],
                    weights=[1 - p_noise, self.p_label_flip, self.p_label_flip],
                ),
                axis=1,
            )
            df_to_annotate["is_noisy_label"] = ~noisy_preference.isna()
            # keeps previously annotated examples when you did not add noise
            df_to_annotate[self.annotation_key] = np.where(
                df_to_annotate["is_noisy_label"],
                noisy_preference,
                df_to_annotate[self.annotation_key],
            )

        # 2. deals with equality
        idcs_is_same_outputs = df_to_annotate["output_1"] == df_to_annotate["output_2"]
        df_to_annotate.loc[idcs_is_same_outputs, self.annotation_key] = 1.5
        # for backward compatibility 0 used to mean same output => replace with 1.5
        df_to_annotate[self.annotation_key] = df_to_annotate[self.annotation_key].replace({0: 1.5})

        return df_to_annotate

    def _filter_annotations_before_storing(self, df_annotated: pd.DataFrame) -> pd.DataFrame:
        if "is_noisy_label" in df_annotated.columns:
            # don't store noisy labels
            df_annotated = df_annotated.query("is_noisy_label == False").drop(columns=["is_noisy_label"])

        df_annotated = super()._filter_annotations_before_storing(df_annotated)
        return df_annotated


# Note: we separate local and json to make it easier to inherit e.g. for having a database version
class PairwiseAnnotator(PairwiseAnnotatorLocal, BaseAnnotatorJSON):
    __doc__ = BaseAnnotatorJSON.__doc__.replace("Base class", "Class") + PAIRWISE_ADDED_DOCSTRING


class SinglePairwiseAnnotator(SingleAnnotator):
    __doc__ = (
        SingleAnnotator.__doc__.replace(
            "A helper class for a single auto annotator.",
            "A helper class for a single pairwise auto annotator.",
        )
        + """
    is_randomize_output_order : bool
        Whether to randomize output_1, output_2 when formatting.
        
    random_seed_keys : str
        The column to use to seed the randomization of output_1, output_2.
    """
    )

    def __init__(
        self,
        *args,
        annotation_column: str = "preference",
        random_seed_column: Sequence[str] = ("instruction",),
        processors_to_kwargs: Optional[dict[str, dict]] = None,
        is_randomize_output_order: bool = True,
        **kwargs,
    ):
        processors_to_kwargs = processors_to_kwargs or {}
        self.is_randomize_output_order = is_randomize_output_order
        if is_randomize_output_order:

            def _fn_replace_if_switch(df: pd.DataFrame) -> pd.DataFrame:
                # applies to annotation_column (preference) 3-x => 2 becomes 1, 1 becomes 2 and everything in between if flaot
                if df.empty or self.annotation_column not in df.columns:
                    return df
                df = df.copy()
                df[self.annotation_column] = df[self.annotation_column].apply(lambda x: 3 - x)
                return df

            # swith output columns by default
            processors_to_kwargs["RandomSwitchTwoColumnsProcessor"] = dict(
                two_columns_to_switch=["output_1", "output_2"],
                fn_replace_if_switch=_fn_replace_if_switch,
                random_seed_columns=random_seed_column,
                _switch_column="is_switched_outputs",  # backward compatibility
            )

        super().__init__(
            *args, annotation_column=annotation_column, processors_to_kwargs=processors_to_kwargs, **kwargs
        )
        self.random_seed_column = list(random_seed_column)

    def _postprocess(self, df_annotated: pd.DataFrame) -> pd.DataFrame:
        df_annotated = super()._postprocess(df_annotated)

        all_values = df_annotated[self.annotation_column]
        all_values = all_values[~all_values.isna()]
        assert all_values.apply(utils.validate_alpacaeval_preference, is_allow_nan=True).all()

        return df_annotated



"""Runs all unit tests for the decoders."""
import math
from types import SimpleNamespace

import pytest

from alpaca_eval.decoders.anthropic import anthropic_completions
from alpaca_eval.decoders.cohere import cohere_completions
from alpaca_eval.decoders.huggingface_api import huggingface_api_completions
from alpaca_eval.decoders.openai import openai_completions

MOCKED_COMPLETION = "Mocked completion text"


@pytest.fixture
def mock_openai_completion():
    # Create a mock Completion object
    completion_mock = dict()
    completion_mock["total_tokens"] = 3
    completion_mock["text"] = MOCKED_COMPLETION
    return completion_mock


def test_openai_completions(mocker, mock_openai_completion):
    # Patch the _openai_completion_helper function to return the mock completion object

    mocker.patch(
        "alpaca_eval.decoders.openai._openai_completion_helper",
        return_value=[mock_openai_completion],
    )
    # use num_procs=1 to avoid issues with pickling
    result = openai_completions(["Prompt 1", "Prompt 2"], "text-davinci-003", batch_size=1, num_procs=1)

    _run_all_asserts_completions(result)


def test_anthropic_completions(mocker):
    mock_response = SimpleNamespace(completion=MOCKED_COMPLETION)

    mocker.patch(
        "alpaca_eval.decoders.anthropic._anthropic_completion_helper",
        return_value=mock_response,
    )
    result = anthropic_completions(["Prompt 1", "Prompt 2"], num_procs=1)
    _run_all_asserts_completions(result)


def test_cohere_completions(mocker):
    mocker.patch(
        "alpaca_eval.decoders.cohere._cohere_completion_helper",
        return_value=["Mocked completion text", 42],
    )
    result = cohere_completions(["Prompt 1", "Prompt 2"], num_procs=1)
    _run_all_asserts_completions(result)


def test_huggingface_api_completions(mocker):
    mocker.patch(
        "alpaca_eval.decoders.huggingface_api.inference_helper",
        return_value=dict(generated_text="Mocked completion text"),
    )
    result = huggingface_api_completions(
        ["Prompt 1", "Prompt 2"],
        model_name="timdettmers/guanaco-33b-merged",
        num_procs=1,
    )
    _run_all_asserts_completions(result)


def _run_all_asserts_completions(result):
    expected_completions = [MOCKED_COMPLETION, MOCKED_COMPLETION]
    assert result["completions"] == expected_completions

    for i in range(len(result["time_per_example"])):
        assert 0 < result["time_per_example"][i] < 1

    assert len(result["price_per_example"]) == 2
    if not math.isnan(result["price_per_example"][0]):
        assert result["price_per_example"][0] == result["price_per_example"][1]
        assert 0 <= result["price_per_example"][0] < 1e-2
    else:
        assert math.isnan(result["price_per_example"][1]) == math.isnan(result["price_per_example"][0])



import logging
from typing import Optional, Sequence

import numpy as np
import torch
import transformers
from peft import PeftModel
from torch.utils.data import Dataset
from tqdm import tqdm
from transformers import AutoModelForCausalLM, AutoTokenizer

from .. import constants, utils

__all__ = ["huggingface_local_completions"]


class ListDataset(Dataset):
    def __init__(self, original_list):
        self.original_list = original_list

    def __len__(self):
        return len(self.original_list)

    def __getitem__(self, i):
        return self.original_list[i]


def huggingface_local_completions(
    prompts: Sequence[str],
    model_name: str,
    do_sample: bool = False,
    batch_size: int = 1,
    model_kwargs=None,
    cache_dir: Optional[str] = constants.DEFAULT_CACHE_DIR,
    is_fast_tokenizer: bool = True,
    adapters_name: Optional[str] = None,
    **kwargs,
) -> dict[str, list]:
    """Decode locally using huggingface transformers pipeline.

    Parameters
    ----------
    prompts : list of str
        Prompts to get completions for.

    model_name : str, optional
        Name of the model (repo on hugging face hub)  to use for decoding.

    do_sample : bool, optional
        Whether to use sampling for decoding.

    batch_size : int, optional
        Batch size to use for decoding. This currently does not work well with to_bettertransformer.

    model_kwargs : dict, optional
        Additional kwargs to pass to from_pretrained.

    cache_dir : str, optional
        Directory to use for caching the model.

    kwargs :
        Additional kwargs to pass to `InferenceApi.__call__`.
    """
    model_kwargs = model_kwargs or {}
    if "device_map" not in model_kwargs:
        model_kwargs["device_map"] = "auto"
    if "torch_dtype" in model_kwargs and isinstance(model_kwargs["torch_dtype"], str):
        model_kwargs["torch_dtype"] = getattr(torch, model_kwargs["torch_dtype"])

    n_examples = len(prompts)
    if n_examples == 0:
        logging.info("No samples to annotate.")
        return []
    else:
        logging.info(f"Using `huggingface_local_completions` on {n_examples} prompts using {model_name}.")

    if not torch.cuda.is_available():
        model_kwargs["load_in_8bit"] = False
        model_kwargs["torch_dtype"] = None

    #  faster but slightly less accurate matrix multiplications
    torch.backends.cuda.matmul.allow_tf32 = torch.backends.cudnn.allow_tf32 = True

    tokenizer = AutoTokenizer.from_pretrained(
        model_name,
        cache_dir=cache_dir,
        padding_side="left",
        use_fast=is_fast_tokenizer,
        **model_kwargs,
    )
    model = AutoModelForCausalLM.from_pretrained(model_name, cache_dir=cache_dir, **model_kwargs).eval()

    if adapters_name:
        logging.info(f"Merging adapter from {adapters_name}.")
        model = PeftModel.from_pretrained(model, adapters_name)
        model = model.merge_and_unload()

    if batch_size == 1:
        try:
            model = model.to_bettertransformer()
        except NotImplementedError:
            pass
        except AttributeError:
            pass

    logging.info(f"Model memory: {model.get_memory_footprint() / 1e9} GB")

    if batch_size > 1:
        # sort the prompts by length so that we don't necessarily pad them by too much
        # save also index to reorder the completions
        original_order, prompts = zip(*sorted(enumerate(prompts), key=lambda x: len(x[1])))
        prompts = list(prompts)

    if not tokenizer.pad_token_id:
        # set padding token if not set
        tokenizer.pad_token_id = tokenizer.eos_token_id
        tokenizer.pad_token = tokenizer.eos_token

    default_kwargs = dict(
        do_sample=do_sample,
        model_kwargs={k: v for k, v in model_kwargs.items() if k != "trust_remote_code"},
        batch_size=batch_size,
    )
    default_kwargs.update(kwargs)
    logging.info(f"Kwargs to completion: {default_kwargs}")
    pipeline = transformers.pipeline(
        task="text-generation",
        model=model,
        tokenizer=tokenizer,
        **default_kwargs,
        trust_remote_code=model_kwargs.get("trust_remote_code", False),
    )

    ## compute and log the time for completions
    prompts_dataset = ListDataset(prompts)
    completions = []

    with utils.Timer() as t:
        for out in tqdm(
            pipeline(
                prompts_dataset,
                return_full_text=False,
                pad_token_id=tokenizer.pad_token_id,
            )
        ):
            completions.append(out[0]["generated_text"])

    logging.info(f"Time for {n_examples} completions: {t}")

    if batch_size > 1:
        # reorder the completions to match the original order
        completions, _ = zip(*sorted(list(zip(completions, original_order)), key=lambda x: x[1]))
        completions = list(completions)

    # local => price is really your compute
    price = [np.nan] * len(completions)
    avg_time = [t.duration / n_examples] * len(completions)

    return dict(completions=completions, price_per_example=price, time_per_example=avg_time)



import abc
import json
import logging
import os
from functools import partial
from pathlib import Path
from typing import Any, Callable, Optional, Sequence, Type, Union

import numpy as np
import pandas as pd

from .. import completion_parsers, constants, processors, utils
from ..decoders import get_fn_completions

CURRENT_DIR = Path(__file__).parent
logging.getLogger().setLevel(logging.INFO)

__all__ = ["BaseAnnotator", "BaseAnnotatorJSON", "SingleAnnotator"]


class BaseAnnotator(abc.ABC):
    """Base class for a pool of annotators.

    Parameters
    ----------
    annotators_config : Path or list of dict, optional
        A dictionary or path to a yaml file containing the configuration for the pool of annotators. If a directory,
        we search for 'configs.yaml' in it. The keys in the first  dictionary should be the annotator's name, and
        the value should be a dictionary of the annotator's configuration which should have the following keys:
        The path is relative to `base_dir` directory.
        - prompt_template (str): a prompt template or path to it. The template should contain placeholders for keys in
            the example dictionary, typically {instruction} and {output_1} {output_2}.
        - fn_completions (str): function in `alpaca_farm.decoders` for completions. Needs to accept as first argument
            `prompts` which is a list of string.
        - completions_kwargs (dict): kwargs for fn_completions. E.g. model_name, max_tokens, temperature,
        tokens_to_avoid
        - fn_completion_parser (str) : Function in `completion_parsers.py` to use for parsing the completions into
        annotations.
        - completion_parser_kwargs (dict) : Kwargs for fn_completion_parser.
        - other kwargs to `SingleAnnotator` such as batch_size

    seed : int, optional
        Seed for the random number generator.

    is_avoid_reannotations : bool, optional
        Whether to avoid re-annotating examples that have already been annotated by the annotator. This will decrease
        cost but can be slightly slower if there are no annotations that can be reused.

    primary_keys : sequence of str, optional
        Keys use to distinguish the example.

    other_output_keys_to_keep : sequence of str, optional
        Other output columns to store besides the annotations.

    other_input_keys_to_keep : sequence of str, optional
        Other columns to keep from the input dataframe besides the primary keys.

    is_store_missing_annotations : bool, optional
        Whether to store missing annotations. If True it avoids trying to reannotate examples that have errors.

    base_dir : Path, optional
        Path to the directory containing the annotators configs. I.e. annotators_config will be relative
        to this directory. If None uses self.DEFAULT_BASE_DIR

    is_raise_if_missing_primary_keys : bool, optional
        Whether to ensure that the primary keys are in the example dictionary. If True, raises an error.

    tmp_missing_annototation : Any, optional
        Temporary value to use for missing annotations when `is_store_missing_annotations` is True.

    annotation_type : type or str, optional
        Type to use for storing the annotations. If None, uses `self.DEFAULT_ANNOTATION_TYPE`.

    is_reapply_parsing : bool, optional
        Whether to reapply the parsing of the completions. This is useful if you want to change the parsing without
        reannotating everything. To be useful you need to have set `is_store_missing_annotations` to True when you
        first annotated.
    """

    DEFAULT_BASE_DIR = constants.EVALUATORS_CONFIG_DIR
    annotator_column = "annotator"
    TMP_MISSING_ANNOTATION = -1
    DEFAULT_ANNOTATION_TYPE = float

    def __init__(
        self,
        primary_keys: Sequence[str],
        annotators_config: Union[utils.AnyPath, list[dict[str, Any]]] = "claude",
        seed: Optional[int] = 0,
        is_avoid_reannotations: bool = True,
        other_output_keys_to_keep: Sequence[str] = ("price_per_example", "time_per_example", "raw_completion"),
        other_input_keys_to_keep: Sequence[str] = (),
        is_store_missing_annotations: bool = True,
        base_dir: Optional[utils.AnyPath] = None,
        is_raise_if_missing_primary_keys: bool = True,
        annotation_type: Optional[Type] = None,
        is_reapply_parsing: bool = False,
    ):
        logging.info(f"Creating the annotator from `{annotators_config}`.")
        self.base_dir = Path(base_dir or self.DEFAULT_BASE_DIR)
        self.seed = seed
        self.is_avoid_reannotations = is_avoid_reannotations
        self.primary_keys = list(primary_keys)
        self.all_keys = self.primary_keys + [self.annotator_column]
        self.other_output_keys_to_keep = list(other_output_keys_to_keep)
        self.other_input_keys_to_keep = list(other_input_keys_to_keep)
        self.other_keys_to_keep = self.other_output_keys_to_keep + self.other_input_keys_to_keep
        self.is_store_missing_annotations = is_store_missing_annotations
        self.is_raise_if_missing_primary_keys = is_raise_if_missing_primary_keys
        if isinstance(annotation_type, str):
            annotation_type = ast.literal_eval(annotation_type)
        self.annotation_type = annotation_type or self.DEFAULT_ANNOTATION_TYPE
        self.is_reapply_parsing = is_reapply_parsing

        self.annotators_config = self._initialize_annotators_config(annotators_config)
        self.annotators = self._initialize_annotators()
        self.df_annotations = None

    ### Abstract methods ###

    #######################
    @property
    def SingleAnnotator(self) -> Type["SingleAnnotator"]:
        """Class to use for each annotator."""
        return SingleAnnotator

    @property
    def available_fields_to_format(self):
        """Fields that can be formatted in the prompt template."""
        return self.all_keys

    @property
    def annotation_key(self) -> str:
        """How to refer to the annotations, this will be the key for annotations in the output."""
        return "annotation"

    @property
    def random_seed_keys(self) -> list[str]:
        """What key / column to seed on for the random generator."""
        return list(self.primary_keys)

    ### Public methods ###
    @property
    def annotator_name(self) -> str:
        return Path(self.annotators_config).parent.name

    def __call__(
        self,
        to_annotate: utils.AnyData,
        chunksize: Optional[int] = 128,
        **decoding_kwargs,
    ) -> list[dict[str, Any]]:
        """Main function for annotating.

        Parameters
        ----------
        to_annotate : list of dict or dataframe
            Examples to annotate. Each dictionary (or row) should contain all of `self.primary_keys`.

        chunksize : int, optional
            The number of rows to annotate at once => ensures that if there is an error, you still get some annotations.

        **decoding_kwargs :
            Additional arguments to pass to `fn_completions`.

        Returns
        -------
        annotated : list of dict
            The annotated examples. Each dict will contain all of `self.primary_keys` and `self.annotation_key`.
        """
        if len(to_annotate) == 0:
            return []

        # note: not ideal potentially doing a lot of dataframe copies. But given that they should be small, ~ok
        df_to_annotate = utils.convert_to_dataframe(to_annotate)
        # make sure primary keys are strings
        for c in self.primary_keys:
            df_to_annotate[c] = df_to_annotate[c].astype(str)

        all_annotated = []
        for df_chunk in utils.dataframe_chunk_generator(df_to_annotate, chunksize, tqdm_desc="Annotation chunk"):
            curr_df_to_annotate = self._preprocess(df_chunk)
            df_annotated = self._annotate(curr_df_to_annotate, **decoding_kwargs)
            annotated = self._postprocess_and_store_(df_annotated, df_chunk)
            all_annotated.extend(annotated)

        return all_annotated

    #######################

    ### Private methods ###
    def _initialize_annotators_config(self, annotators_config):
        # setting it relative to the config directory
        annotators_config = self.base_dir / annotators_config

        if annotators_config.is_dir():
            annotators_config = annotators_config / "configs.yaml"

        return annotators_config

    def _initialize_annotators(self) -> dict[str, "SingleAnnotator"]:
        """Load all the configs and prompts if necessary."""
        annotators_config = utils.load_configs(self.annotators_config)
        try:
            # in case a path is given we make it relative to that path
            base_dir = self.annotators_config.parents[1]
        except:
            base_dir = self.base_dir

        return {
            name: self.SingleAnnotator(
                seed=self.seed,
                base_dir=base_dir,
                annotation_column=self.annotation_key,
                **annotator_config,
            )
            for name, annotator_config in annotators_config.items()
        }

    def _add_missing_primary_keys_(self, df: pd.DataFrame):
        missing_primary_keys = [c for c in self.primary_keys if c not in df.columns]
        if self.is_raise_if_missing_primary_keys:
            if len(missing_primary_keys) > 0:
                raise ValueError(f"Missing primary keys: {missing_primary_keys}")
        else:
            for c in missing_primary_keys:
                df[c] = None

    def _preprocess(self, to_annotate: utils.AnyData) -> pd.DataFrame:
        """Preprocess the examples to annotate. In particular takes care of filtering unnecessary examples."""

        df_to_annotate = utils.convert_to_dataframe(to_annotate)
        self._add_missing_primary_keys_(df_to_annotate)

        # don't remove output keys to keep
        for c in self.other_output_keys_to_keep + [self.annotation_key]:
            if c in df_to_annotate.columns:
                logging.warning(f"{c} column is already in the dataframe. We will overwrite it.")
                df_to_annotate[c] = None

        # remove duplicates because you only need to annotate one of them
        df_to_annotate = df_to_annotate.drop_duplicates(subset=self.primary_keys)

        # set the annotater for each example
        df_to_annotate[self.annotator_column] = df_to_annotate.apply(
            lambda x: utils.random_seeded_choice(
                # we add "annotator" at the beginning to not use the same seed for all tasks
                seed="annotator" + "".join(x[self.random_seed_keys]) + str(self.seed),
                choices=list(self.annotators.keys()),
            ),
            axis=1,
        )

        if self.is_avoid_reannotations:
            df_to_annotate = self._apply_cached_annotations(df_to_annotate)

        return df_to_annotate

    def _annotate(self, df_to_annotate: pd.DataFrame, **decoding_kwargs) -> pd.DataFrame:
        """Annotate the examples."""

        df_annotated = df_to_annotate.copy()
        for annotator in self.annotators.keys():
            # only annotate examples that have not been annotated yet
            curr_idcs = df_to_annotate[self.annotator_column] == annotator
            if self.annotation_key in df_to_annotate.columns:
                curr_idcs &= df_to_annotate[self.annotation_key].isna()

            # drop the output keys that you will be adding
            for k in self.other_output_keys_to_keep:
                if k in df_to_annotate.columns:
                    df_annotated.loc[curr_idcs, k] = None

            logging.info(f"Annotating {curr_idcs.sum()} examples with {annotator}")

            # actual annotation
            columns_to_annotate = self.available_fields_to_format
            if self.is_reapply_parsing:
                # add other_output_keys_to_keep to columns_to_annotate
                columns_to_annotate = columns_to_annotate + [
                    c for c in self.other_output_keys_to_keep if c in df_to_annotate.columns
                ]

            curr_annotated = self.annotators[annotator](
                df_to_annotate.loc[curr_idcs, columns_to_annotate],
                **decoding_kwargs,
            )

            df_annotated = self._merge_annotations(df_annotated, curr_annotated)

        return df_annotated

    def _postprocess_and_store_(
        self,
        df_annotated: pd.DataFrame,
        to_annotate: utils.AnyData,
    ) -> list[dict[str, Any]]:
        """Convert the dataframe into a list of dictionaries to be returned, and store current anntations."""

        df_to_annotate = utils.convert_to_dataframe(to_annotate)
        self._add_missing_primary_keys_(df_to_annotate)

        # select available annotations
        if self.is_store_missing_annotations:
            df_annotated[self.annotation_key] = df_annotated[self.annotation_key].fillna(self.TMP_MISSING_ANNOTATION)
        else:
            df_annotated[self.annotation_key] = df_annotated[self.annotation_key].replace(
                self.TMP_MISSING_ANNOTATION, None
            )

        df_annotated = df_annotated[~df_annotated[self.annotation_key].isna()].copy()

        # try converting to int now that no nan. Note this will only do so if possible
        df_annotated[self.annotation_key] = df_annotated[self.annotation_key].astype(self.annotation_type)

        df_annotated = self._filter_annotations_before_storing(df_annotated)
        self._store_annotations_(df_annotated)

        if self.is_store_missing_annotations:
            # put back None
            df_annotated[self.annotation_key] = df_annotated[self.annotation_key].replace(
                self.TMP_MISSING_ANNOTATION, None
            )

        # need to merge with df_to_annotate in case you dropped duplicates
        on = list(self.primary_keys)
        # keeps columns from both df_to_annotate and df_annotated that are useful
        df_annotated = df_annotated[
            self._get_all_keys_to_keep(list(df_to_annotate.columns) + list(df_annotated.columns))
        ]
        df_to_annotate = df_to_annotate[[c for c in df_to_annotate.columns if c not in df_annotated.columns or c in on]]
        # need to remove all other columns before merging if not you will
        df_annotated = df_to_annotate.merge(df_annotated, on=on, how="outer")

        annotated = df_annotated.to_dict(orient="records")

        return annotated

    def _filter_annotations_before_storing(self, df_annotated: pd.DataFrame) -> pd.DataFrame:
        """Filter annotations before storing them."""
        df_annotated = df_annotated[self._get_all_keys_to_keep(df_annotated.columns)]
        return df_annotated

    def _get_all_keys_to_keep(self, current_columns: Sequence) -> list[str]:
        other_keys_to_keep = [c for c in self.other_keys_to_keep if c in current_columns]
        all_keys_to_keep = self.all_keys + [self.annotation_key] + other_keys_to_keep
        return all_keys_to_keep

    def _apply_cached_annotations(self, df_to_annotate: pd.DataFrame) -> pd.DataFrame:
        """annotate examples with cached annotations"""

        if self.is_store_missing_annotations:
            df_annotations = self.df_annotations
        else:
            # temorarily remove missing annotations from self.df_annotations
            df_annotations = self.df_annotations.query(f"{self.annotation_key} != {self.TMP_MISSING_ANNOTATION}")

        kwargs = {}
        if self.is_reapply_parsing:
            # if you are reapplying parsing then remove the annotation key from the cached annotations
            kwargs = dict(annotation_keys=[])

        df_to_annotate = self._merge_annotations(df_to_annotate, df_annotations, **kwargs)
        return df_to_annotate

    def _store_annotations_(self, df_annotated: pd.DataFrame):
        """Store annotation in memory and on disk"""

        if self.df_annotations is None:
            df_annotations = df_annotated
        else:
            df_annotations = pd.concat([self.df_annotations, df_annotated], axis=0, ignore_index=True)

        self.df_annotations = df_annotations.drop_duplicates(subset=self.all_keys, keep="last")

    def _merge_annotations(
        self,
        df_to_annotate: pd.DataFrame,
        df_partially_annotated: pd.DataFrame,
        annotation_keys: Optional[Sequence] = None,
    ) -> pd.DataFrame:
        """Merge (partial) annotations with the original df to keep the same order and avoid duplicates annotations."""

        if df_partially_annotated is None or df_partially_annotated.empty:
            return df_to_annotate

        other_keys_to_keep = [c for c in self.other_keys_to_keep if c in df_partially_annotated.columns]

        kwargs = dict(
            on=self.all_keys,
            how="left",
            suffixes=("_old", "_new"),
        )

        if annotation_keys is None:
            annotation_keys = [self.annotation_key]

        try:
            df_to_annotate = df_to_annotate.merge(
                df_partially_annotated[self.all_keys + annotation_keys + other_keys_to_keep],
                **kwargs,
            )
        except ValueError:
            # can have merging issues if columns have different dtypes
            df_partially_annotated = df_partially_annotated.astype({k: str for k in self.all_keys})
            df_to_annotate = df_to_annotate.astype({k: str for k in self.all_keys}).merge(
                df_partially_annotated[self.all_keys + annotation_keys + other_keys_to_keep],
                **kwargs,
            )

        # if columns were in both dataframes, try to merge them
        for c in other_keys_to_keep + [self.annotation_key]:
            if f"{c}_old" in df_to_annotate.columns and f"{c}_new" in df_to_annotate.columns:
                df_to_annotate[c] = df_to_annotate[c + "_old"].fillna(df_to_annotate[c + "_new"])
                df_to_annotate = df_to_annotate.drop(columns=[c + "_old", c + "_new"])

        return df_to_annotate

    #######################


class BaseAnnotatorJSON(BaseAnnotator):
    __doc__ = (
        BaseAnnotator.__doc__.replace(
            "Base class for a pool of annotators.",
            "Base class for a pool of annotators with caching to JSON file.",
        )
        + """
    caching_path : Path, optional
        Path to cache the annotations to. If None, will not save the annotations. If the path already exists it will
        load annotations from there.
    """
    )

    def __init__(self, *args, caching_path: Optional[utils.AnyPath] = "auto", **kwargs):
        super().__init__(*args, **kwargs)
        self.caching_path = self._initialize_cache(caching_path)

    def save(self, path: Optional[utils.AnyPath] = None):
        """Save all annotations to json."""

        path = path or self.caching_path
        if path is not None:
            logging.info(f"Saving all annotations to {path}.")
            # to make sure that we don't overwrite the annotations we load again from file (ideally would use a DB)
            self._refresh_annotations_()
            if not self.is_store_missing_annotations:
                self.df_annotations = self.df_annotations[~self.df_annotations[self.annotation_key].isna()]
            self.df_annotations.to_json(path, orient="records", indent=2)

    def load_(self, path: Optional[utils.AnyPath] = None):
        """Load all the annotations from json."""
        path = path or self.caching_path
        if path is not None:
            path = Path(path)
            if path.exists():
                logging.info(f"Loading all annotations from {path}.")
                self.df_annotations = pd.read_json(path, dtype={k: str for k in self.all_keys})

    def _initialize_cache(self, caching_path):
        if caching_path == "auto":
            if isinstance(self.annotators_config, (str, Path, os.PathLike)):
                stem = Path(self.annotators_config).stem
                caching_path = Path(self.annotators_config).parent / f"annotations_seed{self.seed}_{stem}.json"
                logging.info(f"Saving annotations to `{caching_path}`.")
            else:
                logging.warning("caching_path cannot be 'auto' if annotators_config is not a path. Setting to None.")
                caching_path = None
        elif caching_path is not None:
            logging.warning("Saving_path is given but not 'auto', make sure that it's different for different seeds.")

        if caching_path is not None:
            self.load_(caching_path)

        return caching_path

    def _store_annotations_(self, df_annotated_to_store: pd.DataFrame):
        super()._store_annotations_(df_annotated_to_store)
        self.save()

    def _refresh_annotations_(self):
        """Refresh the annotations in memory."""
        curr_df_annotations = self.df_annotations.copy()
        self.load_()
        self.df_annotations = pd.concat(
            [self.df_annotations, curr_df_annotations], axis=0, ignore_index=True
        ).drop_duplicates(subset=self.all_keys, keep="last")


class SingleAnnotator:
    """A helper class for a single auto annotator.

    Parameters
    ----------
    prompt_template : str or path
        A prompt template that will be given to `fn_prompter` or path to those prompts. Path is relative to
        `evaluators_configs/`

    fn_completion_parser : callable or str
        Function that maps (parses) the completion to a list of annotations. If a string, it should be a function in
        `completion_parsers.py` to use for parsing the completions into annotations. For each completion, the number of
        annotations (lenght of list) should be equal to the batch_size if not we set all the annotations in that batch
        to NaN.

    completion_parser_kwargs : dict
        Kwargs for fn_completion_parser.

    fn_completions : callable or str
        Function in `decoders.py` to use for decoding the output.

    completions_kwargs : dict
        kwargs for fn_completions. E.g. model_name, max_tokens, temperature, top_p, top_k, stop_seq.

    is_shuffle : bool
        Whether to shuffle the order of the examples before making the prompt. Useful if batch_size > 1.

    seed : int
        Seed for randomization.

    batch_size : int
        Number of examples that will be added in a single prompt.

    base_dir : Path, optional
        Path to the directory containing the annotators configs. I.e. annotators_config will be relative
        to this directory.

    annotation_column : str, optional
        Name of the annotation column in the output dataframe.

    is_store_raw_completions : bool, optional
        Whether to store raw completions at `"raw_completion"` column in the output dataframe. Note that raw_completion
        will not be modified by the postprocessors. E.g. if we switch the columns output_1 and output_2 in the prompt
        then the raw completion will show the switched order, which makes interpretation harder. This should
        nevertheless not be an issue when using reapply_parsing because of seeding.

    processors_to_kwargs : Sequence[dict(str, dict)], optional
        A dictionary of BaseProcessor objects to apply for preprocessing the  dataframe before making the prompts and
        prostprocessing after anntoations. The key should be the names of the BaseProcessor objectsto use in
        `processors.py` the values are the kwargs for the constructor of the Processor. Order matters.

    is_add_default_processors : bool, optional
        Whether to add the default processors to the list of processors.

    completion_key : str, optional
        Key of the output of `fn_completions` to use for parsing the completions into annotations.

    is_json_load_raw_completions : bool, optional
        Whether to try to load the raw completions into a json. If exceptionthen will return the raw completions as is.
    """

    def __init__(
        self,
        prompt_template: utils.AnyPath,
        fn_completion_parser: Optional[Union[Callable, str]] = "regex_parser",
        completion_parser_kwargs: Optional[dict[str, Any]] = None,
        fn_completions: Union[Callable, str] = "openai_completions",
        completions_kwargs: Optional[dict[str, Any]] = None,
        is_shuffle: bool = True,
        seed: Optional[int] = 123,
        batch_size: int = 1,
        base_dir: utils.AnyPath = constants.EVALUATORS_CONFIG_DIR,
        annotation_column: str = "annotation",
        is_store_raw_completions: bool = True,
        processors_to_kwargs: Optional[dict[str, dict]] = None,
        is_add_default_processors: bool = True,
        completion_key: str = "completions",
        is_json_load_raw_completions: bool = False,
    ):
        self.base_dir = Path(base_dir)
        self.prompt_template = self._get_prompt_template(prompt_template)

        if fn_completion_parser is None:
            fn_completion_parser = lambda x: [x]
        elif isinstance(fn_completion_parser, str):
            fn_completion_parser = self._search_fn_completion_parser(fn_completion_parser)
        completion_parser_kwargs = completion_parser_kwargs or {}
        self.fn_completion_parser = partial(fn_completion_parser, **completion_parser_kwargs)

        self.fn_completions = get_fn_completions(fn_completions)
        self.completions_kwargs = completions_kwargs or {}
        self.seed = seed
        self.is_shuffle = is_shuffle
        self.batch_size = batch_size
        self.annotation_column = annotation_column
        self.completion_column = "raw_completion" if is_store_raw_completions else None
        self.is_json_load_raw_completions = is_json_load_raw_completions

        self.is_add_default_processors = is_add_default_processors
        self.processors = []
        self.completion_key = completion_key
        processors_to_kwargs = processors_to_kwargs or {}
        if (
            batch_size > 1
            and self.is_add_default_processors
            and "PaddingForBatchesProcessor" not in processors_to_kwargs
        ):
            processors_to_kwargs["PaddingForBatchesProcessor"] = {
                "batch_size": batch_size,
                "padding_example": utils.DUMMY_EXAMPLE,
            }
        for processor, processor_kwargs in processors_to_kwargs.items():
            processor_kwargs["seed"] = self.seed
            Processor = self._search_processor(processor)
            self.processors += [Processor(**processor_kwargs)]

    ### Public methods ###
    def __call__(self, df_to_annotate: pd.DataFrame, **decoding_kwargs) -> pd.DataFrame:
        """Annotates the given examples.

        Parameters
        ----------
        df_to_annotate : pd.DataFrame
            Examples to annotate

        decoding_kwargs :
            Additional arguments to pass to `fn_completions`.
        """
        df_to_annotate = df_to_annotate.copy()  # avoid in place modifications

        if df_to_annotate.empty:
            df_to_annotate[self.annotation_column] = []
            return df_to_annotate

        df_to_annotate = self._preprocess(df_to_annotate)

        # the following only reapplies the parsing in case you already stored the raw completions. requires batch_size=1
        if self.completion_column in df_to_annotate.columns and self.batch_size == 1:
            # keep only the rows that have not been annotated yet
            main_df_to_annotate = df_to_annotate
            idx_not_completed = df_to_annotate[self.completion_column].isna()
            df_to_annotate = df_to_annotate[idx_not_completed].copy()

        if not df_to_annotate.empty:
            # prompts and completions here will not be the same length as the dataframe due to batching
            prompts, df_to_annotate = self._make_prompts(df_to_annotate)
            completions = self.fn_completions(prompts=prompts, **self.completions_kwargs, **decoding_kwargs)

            for k, v in completions.items():
                if k != "completions":
                    if self.batch_size != 1 and (len(df_to_annotate) == len(v) * self.batch_size):
                        v = [el for el in v for _ in range(self.batch_size)]
                    df_to_annotate[k] = v
                    if "per_example" in k:
                        df_to_annotate[k] = df_to_annotate[k] / self.batch_size

        # the following is only needed if you want to only reapply the parsing
        if self.completion_column in df_to_annotate.columns:
            if not df_to_annotate.empty:
                df_to_annotate[self.completion_column] = completions[self.completion_key]  # only works for bs 1
            main_df_to_annotate[idx_not_completed] = df_to_annotate  # puts back all the new completions
            df_to_annotate = main_df_to_annotate
            completions_to_parse = df_to_annotate[self.completion_column]
        else:
            completions_to_parse = completions[self.completion_key]

        # note: reparsing only works if you use the same completion_key
        annotations_to_save, completions_to_save = self._parse_completions(completions=completions_to_parse)
        df_to_annotate[self.annotation_column] = annotations_to_save
        if self.completion_column is not None:
            df_to_annotate[self.completion_column] = completions_to_save

        df_annotated = self._postprocess(df_to_annotate)

        return df_annotated

    ######################

    ### Private methods ###
    def _search_fn_completion_parser(self, name: str) -> Callable:
        """Search for a completion parser by name."""
        return utils.get_module_attribute(completion_parsers, name)

    def _search_processor(self, name: Union[str, Type["processors.BaseProcessor"]]) -> Type["processors.BaseProcessor"]:
        """Search for a Processor class by name."""
        if isinstance(name, str):
            return utils.get_module_attribute(processors, name)
        else:
            assert issubclass(name, processors.BaseProcessor)
            return name

    def _get_prompt_template(self, prompt_template: utils.AnyPath):
        return utils.read_or_return(self.base_dir / prompt_template)

    def _make_prompts(
        self, df_to_annotate: pd.DataFrame, prompt_template: Optional[str] = None
    ) -> tuple[list[str], pd.DataFrame]:
        """Make all the prompts for the given examples.

        Parameters
        ----------
        df_to_annotate : pd.DataFrame
            Examples to annotate

        prompt_template : str
            Template to use for the prompt. If None, use the one from the constructor.

        Returns
        -------
        prompts : list[str]
            Formatted prompts for the given examples.

        df_to_annotate : pd.DataFrame
            Examples to annotate in the same order as prompts.
        """
        if prompt_template is None:
            prompt_template = self.prompt_template
        return utils.make_prompts(df=df_to_annotate, template=prompt_template, batch_size=self.batch_size)

    def _preprocess(self, df_to_annotate: pd.DataFrame) -> pd.DataFrame:
        """Preprocess the examples before annotating. In particular, takes care of all the randomization."""

        for processor in self.processors:
            df_to_annotate = processor.preprocess(df_to_annotate)

        if self.is_shuffle:
            df_to_annotate = df_to_annotate.sample(frac=1, random_state=self.seed)

        return df_to_annotate

    def _parse_completions(self, completions: list[str]) -> tuple[list[Any], list[Any]]:
        """Converts the completions into annotations."""
        all_annotations = []
        all_completions = []
        for completion in completions:
            try:
                batch_annotations = self.fn_completion_parser(completion)
                batch_annotations = list(batch_annotations)

                if len(batch_annotations) != self.batch_size:
                    logging.warning(
                        f"Found {len(batch_annotations)} annotations in:'''\n{completion}\n''' but expected"
                        f" {self.batch_size}. We are setting all annotations to None."
                    )
                    batch_annotations = [None] * self.batch_size

            except Exception as e:
                logging.exception(f"Error while parsing completion: '''\n{completion}\n'''")
                batch_annotations = [None] * self.batch_size

            all_annotations += batch_annotations

            if self.is_json_load_raw_completions:
                try:
                    completion = json.loads(completion)
                except:
                    pass

            all_completions += [completion] * self.batch_size
        return all_annotations, all_completions

    def _postprocess(self, df_annotated: pd.DataFrame) -> pd.DataFrame:
        """Postprocess the annotated examples."""

        arr_is_na = df_annotated[self.annotation_column].isna()
        if arr_is_na.any():
            logging.warning(
                f"{arr_is_na.sum().item()} samples had no auto annotation. We are filtering them for now. "
                f"If you are using chain of thought it might be that max_tokens limit is too low. "
            )
            df_annotated = df_annotated[~arr_is_na]

        for processor in self.processors[::-1]:  # postprocess in reverted order => no interactions between processors
            df_annotated = processor.postprocess(df_annotated)

        return df_annotated

    #######################



import copy
import functools
import logging
import multiprocessing
import random
from typing import Optional, Sequence, Tuple

import cohere
import tqdm
from cohere import CohereError

from .. import constants, utils

__all__ = ["cohere_completions"]


def cohere_completions(
    prompts: Sequence[str],
    model_name="command",
    num_procs: int = 5,
    **decoding_kwargs,
) -> dict[str, list]:
    """Decode with Cohere API.

    Parameters
    ----------
    prompts : list of str
        Prompts to get completions for.

    model_name : str, optional
        Name of the model to use for decoding.

    num_procs : int, optional
        Number of parallel processes to use for decoding.

    decoding_kwargs :
        Additional kwargs to pass to `cohere.Client.generation`.
    """
    n_examples = len(prompts)
    if n_examples == 0:
        logging.info("No samples to annotate.")
        return []
    else:
        logging.info(f"Using `cohere_completions` on {n_examples} prompts using {model_name}.")

    kwargs = dict(model=model_name, **decoding_kwargs)
    logging.info(f"Kwargs to completion: {kwargs}")

    with utils.Timer() as t:
        if num_procs == 1:
            completions_and_token_counts = [
                _cohere_completion_helper(prompt, **kwargs) for prompt in tqdm.tqdm(prompts, desc="prompts")
            ]
        else:
            with multiprocessing.Pool(num_procs) as p:
                partial_completion_helper = functools.partial(_cohere_completion_helper, **kwargs)
                completions_and_token_counts = list(
                    tqdm.tqdm(
                        p.imap(partial_completion_helper, prompts),
                        desc="prompts",
                        total=len(prompts),
                    )
                )
    logging.info(f"Completed {n_examples} examples in {t}.")
    completions, num_tokens = zip(*completions_and_token_counts)
    price_per_token = 0.000015  # cohere charges $0.000015 per token.
    price_per_example = [price_per_token * n for n in num_tokens]
    avg_time = [t.duration / n_examples] * len(completions)

    return dict(completions=list(completions), price_per_example=price_per_example, time_per_example=avg_time)


def _cohere_completion_helper(
    prompt: str,
    cohere_api_keys: Optional[Sequence[str]] = (constants.COHERE_API_KEY,),
    max_tokens: Optional[int] = 1000,
    temperature: Optional[float] = 0.7,
    max_tries=5,
    **kwargs,
) -> Tuple[str, int]:
    cohere_api_key = random.choice(cohere_api_keys)
    client = cohere.Client(cohere_api_key)

    kwargs.update(dict(max_tokens=max_tokens, temperature=temperature))
    curr_kwargs = copy.deepcopy(kwargs)

    for trynum in range(max_tries):  # retry errors
        try:
            response = client.generate(prompt=prompt, return_likelihoods="ALL", **curr_kwargs)
            text = response[0].text
            num_tokens = len(response[0].token_likelihoods)

            if text == "":
                raise CohereError("Empty string response")

            return text, num_tokens

        except CohereError as e:
            logging.info(f"Try #{trynum+1}/{max_tries}: Error running prompt {repr(prompt)}: {e}")

    return " ", 0  # placeholder response for errors, doesn't allow empty string



from unittest.mock import MagicMock

import pandas as pd
import pytest

from alpaca_eval.annotators import PairwiseAnnotator, SinglePairwiseAnnotator


@pytest.fixture
def df_to_annotate():
    return pd.DataFrame(
        {
            "instruction": ["2+2", "1+1", "2+3"],
            "output_1": ["4", "2", "5"],
            "output_2": ["5", "1", "4"],
        }
    )


@pytest.fixture
def expected_annotations():
    return [
        {
            "instruction": "2+2",
            "output_1": "4",
            "output_2": "5",
            "annotator": "test",
            "preference": 1,
        },
        {
            "instruction": "1+1",
            "output_1": "2",
            "output_2": "1",
            "annotator": "test",
            "preference": 2,
        },
        {
            "instruction": "2+3",
            "output_1": "5",
            "output_2": "4",
            "annotator": "test",
            "preference": 1,
        },
    ]


@pytest.fixture
def single_annotator():
    return SinglePairwiseAnnotator(
        prompt_template="text_davinci_003/basic_prompt.txt",
        completion_parser_kwargs=dict(outputs_to_match={1: r"(?:^|\n) ?Output \(a\)", 2: "(?:^|\n) ?Output \(b\)"}),
        is_randomize_output_order=False,
        is_shuffle=False,
        is_store_raw_completions=False,
    )


def test_single_annotator(single_annotator, df_to_annotate):
    # Create a sample DataFrame for testing
    parsable_completions = ["Output (a)", "Output (b)"]
    completions = parsable_completions + ["not parsable"]  # add an example that can't be parsed
    single_annotator.fn_completions = MagicMock(return_value={"completions": completions})
    # set a completion_column => store it
    single_annotator.completion_column = "completions"

    # Call the preprocess method
    df_annotated = single_annotator(df_to_annotate)

    assert df_annotated["preference"].tolist() == [1, 2]
    assert df_annotated["instruction"].tolist() == ["2+2", "1+1"]
    assert set(df_annotated.columns.tolist()) == {
        "instruction",
        "output_1",
        "output_2",
        "preference",
        "completions",
    }
    # check that you also save the completions.
    assert df_annotated["completions"].tolist() == parsable_completions


@pytest.fixture
def pairwise_annotator(tmp_path):
    return PairwiseAnnotator(
        annotators_config="test",
        caching_path=tmp_path / "cache_{seed}.json",
    )


def _get_mock_annotator(annotator, preference=None):
    def annotate(df):
        df = df.copy()
        if preference is None:
            df["preference"] = [1] * len(df)
        else:
            df["preference"] = preference
        return df

    mock_function = MagicMock(side_effect=annotate)
    for a in annotator.annotators.keys():
        annotator.annotators[a] = mock_function

    return annotator, mock_function


def test_annotate_pairs(pairwise_annotator, df_to_annotate, expected_annotations):
    pairwise_annotator, mock_function = _get_mock_annotator(pairwise_annotator, preference=[1, 2, 1])

    # Call the annotate_pairs method and assert output
    annotated = pairwise_annotator.annotate_pairs(df_to_annotate)
    assert annotated == expected_annotations
    args, kwargs = mock_function.call_args
    assert len(args[0]) == 3

    # now try again with caching => should not be annotating anything
    annotated = pairwise_annotator.annotate_pairs(df_to_annotate)
    assert annotated == expected_annotations
    args, kwargs = mock_function.call_args
    assert len(args[0]) == 0

    # now try again without caching => should be annotating again
    pairwise_annotator.is_avoid_reannotations = False
    annotated = pairwise_annotator.annotate_pairs(df_to_annotate)
    assert annotated == expected_annotations
    args, kwargs = mock_function.call_args
    assert len(args[0]) == 3


def test_annotate_samples(pairwise_annotator):
    # this has a list of outputs and you will samples pairs from it
    potential_outputs = ["5", "1", "4"]
    outputs_to_annotate = [
        {
            "instruction": "2+2",
            "output": potential_outputs,
        }
    ]

    pairwise_annotator, mock_function = _get_mock_annotator(pairwise_annotator)

    def run_all_tests(annotated_list_of_dict):
        # run tests for each row => use dataframe
        assert isinstance(annotated_list_of_dict, list)
        assert len(annotated_list_of_dict[0]), dict
        df_annotated = pd.DataFrame(annotated_list_of_dict)
        assert (df_annotated["instruction"] == "2+2").all()
        assert df_annotated["output_1"].isin(potential_outputs).all()
        assert df_annotated["output_2"].isin(potential_outputs).all()
        assert (df_annotated["output_1"] != df_annotated["output_2"]).all()
        assert (df_annotated["preference"] == 1).all()

    # let's sample one pair from the list of outputs (is_unique_instructions=True)
    annotated = pairwise_annotator.annotate_samples(outputs_to_annotate, is_unique_instructions=True)
    assert len(annotated) == 1
    run_all_tests(annotated)

    # same but one pair per output (is_unique_instructions=False)
    annotated = pairwise_annotator.annotate_samples(outputs_to_annotate, is_unique_instructions=False)
    assert len(annotated) == 3
    run_all_tests(annotated)


def test_annotate_head2head(pairwise_annotator, df_to_annotate, expected_annotations):
    pairwise_annotator, mock_function = _get_mock_annotator(pairwise_annotator, preference=[1, 2, 1])

    df_1 = df_to_annotate.drop(columns="output_2").rename(columns={"output_1": "output"})
    df_2 = df_to_annotate.drop(columns="output_1").rename(columns={"output_2": "output"})
    annotated = pairwise_annotator.annotate_head2head(df_1, df_2, is_ordered=True)

    assert annotated == expected_annotations
    args, kwargs = mock_function.call_args
    assert len(args[0]) == 3

    # let's change the order of one dataframe and see if results are the same
    df_2 = df_2.iloc[::-1]
    annotated = pairwise_annotator.annotate_head2head(df_1, df_2, is_ordered=False)

    assert annotated == expected_annotations



import pandas as pd
import pytest

from alpaca_eval import analyze
from alpaca_eval.analyze import SCORING_RULES


@pytest.fixture
def analyzer():
    return analyze.Analyzer(
        n_annotators=4, gold_crossannotations=RECORDS, gold_annotations=None, scoring_rule="zero_one"
    )


def test_agreement_of_annotations(analyzer):
    df_crossannotations = analyzer.df_gold_crossannotations.head(8).copy()
    # index is 0,1,2,3,0,1,2,3
    df_crossannotations["preference"] = [1] * 4 + [2, 2, 2, 1]
    agreement = analyzer.agreement_of_annotations(
        df_crossannotations,
        annotations_2=None,
        n_majority_vote_1=1,
        n_majority_vote_2=1,
    )
    # for first example is always correct => 1. for second example, agreement is 2/3, 2/3, 2/3, 0 => 6/12
    # in total mean(1,1/2) = 0.75
    assert agreement["score"] == pytest.approx(0.75)
    assert agreement["sem_samples"] == pytest.approx(0.25)
    assert agreement["counts"] == pytest.approx(2)
    assert agreement["sem_annotators"] == pytest.approx(0.07537783614444091)

    # now with 3 annotators
    kwargs_3_annotators = dict(
        annotations_1=df_crossannotations,
        annotations_2=None,
        n_majority_vote_1=1,
        n_majority_vote_2=3,
    )
    agreement = analyzer.agreement_of_annotations(**kwargs_3_annotators)
    # for first example is always correct => 1. for second example, agreement is 1, 1, 1, 0 => 3/4
    # in total mean(1,3/4) = 0.875. Where the 1 come from the mode
    assert agreement["score"] == pytest.approx(0.875)
    assert agreement["sem_samples"] == pytest.approx(0.125)
    assert agreement["counts"] == pytest.approx(2)
    assert agreement["sem_annotators"] == pytest.approx(0.125)

    # now with scoring rule "absolute", which should give the same result as predictions are discrete
    analyzer.scoring_rule = SCORING_RULES["absolute"]()
    agreement = analyzer.agreement_of_annotations(**kwargs_3_annotators)
    assert agreement["score"] == pytest.approx(0.875)

    # now change a little the preferences
    df_crossannotations["preference"] = [1] * 4 + [2, 2, 2, 1.5]
    agreement = analyzer.agreement_of_annotations(**kwargs_3_annotators)
    # for first example is always correct => 1. for second example, agreement is 1,1,1,0.5 => 0.875
    # in total mean(1,0.875) = 0.9375.
    assert agreement["score"] == pytest.approx(0.9375)

    df_crossannotations["preference"] = [1] * 4 + [1.9, 1.8, 1.7, 1.6]
    agreement = analyzer.agreement_of_annotations(**kwargs_3_annotators)
    # for first example is always correct => 1. for second example, agreement is 0.8,0.9,0.9,0.8 => 0.85
    # in total mean(1,0.85) = 0.925.
    assert agreement["score"] == pytest.approx(0.925)


def test_get_length_biases(analyzer):
    # Create a sample DataFrame for testing
    df = pd.DataFrame(
        {
            "instruction": ["A", "B", "C"],
            "output_1": ["long output", "short", "much longer output" * 30],
            "output_2": ["short", "somewhat short", "shorter output"],
            "preference": [2, 1, 1],
        }
    )

    # Test the get_length_biases method
    result = analyzer.get_length_biases(df)
    assert result["probability_prefer_longer"] == pytest.approx(1.0)
    assert result["percentage_longer"] == pytest.approx(12.127705627705629)


def test_get_list_biases(analyzer):
    # Create a sample DataFrame for testing
    df = pd.DataFrame(
        {
            "instruction": ["A", "B", "C"],
            "output_1": ["apple", "- apple\n - banana", "apple"],
            "output_2": ["1. apple\n 2. banana", "apple", "a. apple\n b. banana"],
            "preference": [1, 1, 2],
        }
    )

    # Test the get_list_biases method
    result = analyzer.get_list_biases(df)
    assert result["probability_prefer_list"] == pytest.approx(0.6666666667)


RECORDS = [
    {
        "instruction": "The sentence you are given might be too wordy, complicated, or unclear. Rewrite the "
        "sentence and make your writing clearer by keeping it concise. Whenever possible, "
        "break complex sentences into multiple sentences and eliminate unnecessary words.\n\nIf "
        "you have any questions about my rate or if you find it necessary to increase or "
        "decrease the scope for this project, please let me know.",
        "output_1": "If you have questions about my rate or need to modify the scope of this project, "
        "please let me know.",
        "output_2": "If you have any questions about my rate or need to adjust the scope for this project, "
        "please let me know.",
        "preference": 1,
        "annotator_index": 15,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 0,
        "n_annotated": 4,
    },
    {
        "instruction": "The sentence you are given might be too wordy, complicated, or unclear. Rewrite the "
        "sentence and make your writing clearer by keeping it concise. Whenever possible, "
        "break complex sentences into multiple sentences and eliminate unnecessary words.\n\nIf "
        "you have any questions about my rate or if you find it necessary to increase or "
        "decrease the scope for this project, please let me know.",
        "output_1": "If you have questions about my rate or need to modify the scope of this project, "
        "please let me know.",
        "output_2": "If you have any questions about my rate or need to adjust the scope for this project, "
        "please let me know.",
        "preference": 1,
        "annotator_index": 0,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 1,
        "n_annotated": 4,
    },
    {
        "instruction": "The sentence you are given might be too wordy, complicated, or unclear. Rewrite the "
        "sentence and make your writing clearer by keeping it concise. Whenever possible, "
        "break complex sentences into multiple sentences and eliminate unnecessary words.\n\nIf "
        "you have any questions about my rate or if you find it necessary to increase or "
        "decrease the scope for this project, please let me know.",
        "output_1": "If you have questions about my rate or need to modify the scope of this project, "
        "please let me know.",
        "output_2": "If you have any questions about my rate or need to adjust the scope for this project, "
        "please let me know.",
        "preference": 1,
        "annotator_index": 9,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 2,
        "n_annotated": 4,
    },
    {
        "instruction": "The sentence you are given might be too wordy, complicated, or unclear. Rewrite the "
        "sentence and make your writing clearer by keeping it concise. Whenever possible, "
        "break complex sentences into multiple sentences and eliminate unnecessary words.\n\nIf "
        "you have any questions about my rate or if you find it necessary to increase or "
        "decrease the scope for this project, please let me know.",
        "output_1": "If you have questions about my rate or need to modify the scope of this project, "
        "please let me know.",
        "output_2": "If you have any questions about my rate or need to adjust the scope for this project, "
        "please let me know.",
        "preference": 2,
        "annotator_index": 7,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 3,
        "n_annotated": 4,
    },
    {
        "instruction": "Analyze the word choice, phrasing, punctuation, and capitalization in the given email. "
        "How may the writer of this email sound to the reader? These tones include "
        "Disheartening, Accusatory, Worried, Curious, Surprised, Disapproving, Unassuming, "
        "Formal, Assertive, Confident, Appreciative, Concerned, Sad, Informal, Regretful, "
        "Encouraging, Egocentric, Joyful, Optimistic, and Excited.\n\nHi Jen, \nI hope you're "
        "well. Can we catch up today? I'd appreciate your input on my presentation for "
        "tomorrow's meeting. I'd especially love it if you could double-check the sales numbers "
        "with me. There's a coffee in it for you!",
        "output_1": "The writer of this email likely sounds Appreciative, Encouraging, and Optimistic.",
        "output_2": "The tone of the email is mostly informal, with a hint of laughter and enthusiasm.",
        "preference": 1,
        "annotator_index": 10,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 0,
        "n_annotated": 4,
    },
    {
        "instruction": "Analyze the word choice, phrasing, punctuation, and capitalization in the given email. "
        "How may the writer of this email sound to the reader? These tones include "
        "Disheartening, Accusatory, Worried, Curious, Surprised, Disapproving, Unassuming, "
        "Formal, Assertive, Confident, Appreciative, Concerned, Sad, Informal, Regretful, "
        "Encouraging, Egocentric, Joyful, Optimistic, and Excited.\n\nHi Jen, \nI hope you're "
        "well. Can we catch up today? I'd appreciate your input on my presentation for "
        "tomorrow's meeting. I'd especially love it if you could double-check the sales numbers "
        "with me. There's a coffee in it for you!",
        "output_1": "The writer of this email likely sounds Appreciative, Encouraging, and Optimistic.",
        "output_2": "The tone of the email is mostly informal, with a hint of laughter and enthusiasm.",
        "preference": 2,
        "annotator_index": 15,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 1,
        "n_annotated": 4,
    },
    {
        "instruction": "Analyze the word choice, phrasing, punctuation, and capitalization in the given email. "
        "How may the writer of this email sound to the reader? These tones include "
        "Disheartening, Accusatory, Worried, Curious, Surprised, Disapproving, Unassuming, "
        "Formal, Assertive, Confident, Appreciative, Concerned, Sad, Informal, Regretful, "
        "Encouraging, Egocentric, Joyful, Optimistic, and Excited.\n\nHi Jen, \nI hope you're "
        "well. Can we catch up today? I'd appreciate your input on my presentation for "
        "tomorrow's meeting. I'd especially love it if you could double-check the sales numbers "
        "with me. There's a coffee in it for you!",
        "output_1": "The writer of this email likely sounds Appreciative, Encouraging, and Optimistic.",
        "output_2": "The tone of the email is mostly informal, with a hint of laughter and enthusiasm.",
        "preference": 1,
        "annotator_index": 0,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 2,
        "n_annotated": 4,
    },
    {
        "instruction": "Analyze the word choice, phrasing, punctuation, and capitalization in the given email. "
        "How may the writer of this email sound to the reader? These tones include "
        "Disheartening, Accusatory, Worried, Curious, Surprised, Disapproving, Unassuming, "
        "Formal, Assertive, Confident, Appreciative, Concerned, Sad, Informal, Regretful, "
        "Encouraging, Egocentric, Joyful, Optimistic, and Excited.\n\nHi Jen, \nI hope you're "
        "well. Can we catch up today? I'd appreciate your input on my presentation for "
        "tomorrow's meeting. I'd especially love it if you could double-check the sales numbers "
        "with me. There's a coffee in it for you!",
        "output_1": "The writer of this email likely sounds Appreciative, Encouraging, and Optimistic.",
        "output_2": "The tone of the email is mostly informal, with a hint of laughter and enthusiasm.",
        "preference": 1,
        "annotator_index": 4,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 3,
        "n_annotated": 4,
    },
    {
        "instruction": "Rewrite the given text and correct grammar, spelling, and punctuation errors.\n\nIf "
        "you'd told me year ago that today I would finish a marathon, I would of laughed. Your "
        "support had a huge affect on me!",
        "output_1": "If you had told me a year ago that today I would finish a marathon, I would have laughed. "
        "Your support had a huge effect on me!",
        "output_2": "If you had told me a year ago that today I would finish a marathon, I would have laughed. "
        "Your support had a huge affect on me!",
        "preference": 1,
        "annotator_index": 7,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 0,
        "n_annotated": 4,
    },
    {
        "instruction": "Rewrite the given text and correct grammar, spelling, and punctuation errors.\n\nIf "
        "you'd told me year ago that today I would finish a marathon, I would of laughed. Your "
        "support had a huge affect on me!",
        "output_1": "If you had told me a year ago that today I would finish a marathon, I would have laughed. "
        "Your support had a huge effect on me!",
        "output_2": "If you had told me a year ago that today I would finish a marathon, I would have laughed. "
        "Your support had a huge affect on me!",
        "preference": 1,
        "annotator_index": 8,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 1,
        "n_annotated": 4,
    },
    {
        "instruction": "Rewrite the given text and correct grammar, spelling, and punctuation errors.\n\nIf "
        "you'd told me year ago that today I would finish a marathon, I would of laughed. Your "
        "support had a huge affect on me!",
        "output_1": "If you had told me a year ago that today I would finish a marathon, I would have laughed. "
        "Your support had a huge effect on me!",
        "output_2": "If you had told me a year ago that today I would finish a marathon, I would have laughed. "
        "Your support had a huge affect on me!",
        "preference": 1,
        "annotator_index": 11,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 2,
        "n_annotated": 4,
    },
    {
        "instruction": "Rewrite the given text and correct grammar, spelling, and punctuation errors.\n\nIf "
        "you'd told me year ago that today I would finish a marathon, I would of laughed. Your "
        "support had a huge affect on me!",
        "output_1": "If you had told me a year ago that today I would finish a marathon, I would have laughed. "
        "Your support had a huge effect on me!",
        "output_2": "If you had told me a year ago that today I would finish a marathon, I would have laughed. "
        "Your support had a huge affect on me!",
        "preference": 2,
        "annotator_index": 1,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 3,
        "n_annotated": 4,
    },
    {
        "instruction": "You are given a paper citation, convert it to the requested citation "
        "style.\n\nChicago: Vaswani, Ashish, Shazeer, Noam, Parmar, Niki, Uszkoreit, Jakob, "
        'Jones, Llion, Gomez, Aidan N., Kaiser, Lukasz, and Illia Polosukhin. "Attention Is All '
        'You Need." arXiv, (2017). https://doi.org/10.48550/arXiv.1706.03762.\nMLA:',
        "output_1": 'Vaswani, Ashish, et al. "Attention Is All You Need." arXiv, '
        "2017. https://doi.org/10.48550/arXiv.1706.03762.",
        "output_2": 'Vaswani, Ashish et al. "Attention Is All You Need." arXiv, 2017, '
        "https://doi.org/10.48550/arXiv.1706.03762. MLA.",
        "preference": 1,
        "annotator_index": 2,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 0,
        "n_annotated": 4,
    },
    {
        "instruction": "You are given a paper citation, convert it to the requested citation "
        "style.\n\nChicago: Vaswani, Ashish, Shazeer, Noam, Parmar, Niki, Uszkoreit, Jakob, "
        'Jones, Llion, Gomez, Aidan N., Kaiser, Lukasz, and Illia Polosukhin. "Attention Is All '
        'You Need." arXiv, (2017). https://doi.org/10.48550/arXiv.1706.03762.\nMLA:',
        "output_1": 'Vaswani, Ashish, et al. "Attention Is All You Need." arXiv, '
        "2017. https://doi.org/10.48550/arXiv.1706.03762.",
        "output_2": 'Vaswani, Ashish et al. "Attention Is All You Need." arXiv, 2017, '
        "https://doi.org/10.48550/arXiv.1706.03762. MLA.",
        "preference": 1,
        "annotator_index": 0,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 1,
        "n_annotated": 4,
    },
    {
        "instruction": "You are given a paper citation, convert it to the requested citation "
        "style.\n\nChicago: Vaswani, Ashish, Shazeer, Noam, Parmar, Niki, Uszkoreit, Jakob, "
        'Jones, Llion, Gomez, Aidan N., Kaiser, Lukasz, and Illia Polosukhin. "Attention Is All '
        'You Need." arXiv, (2017). https://doi.org/10.48550/arXiv.1706.03762.\nMLA:',
        "output_1": 'Vaswani, Ashish, et al. "Attention Is All You Need." arXiv, '
        "2017. https://doi.org/10.48550/arXiv.1706.03762.",
        "output_2": 'Vaswani, Ashish et al. "Attention Is All You Need." arXiv, 2017, '
        "https://doi.org/10.48550/arXiv.1706.03762. MLA.",
        "preference": 1,
        "annotator_index": 15,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 2,
        "n_annotated": 4,
    },
    {
        "instruction": "You are given a paper citation, convert it to the requested citation "
        "style.\n\nChicago: Vaswani, Ashish, Shazeer, Noam, Parmar, Niki, Uszkoreit, Jakob, "
        'Jones, Llion, Gomez, Aidan N., Kaiser, Lukasz, and Illia Polosukhin. "Attention Is All '
        'You Need." arXiv, (2017). https://doi.org/10.48550/arXiv.1706.03762.\nMLA:',
        "output_1": 'Vaswani, Ashish, et al. "Attention Is All You Need." arXiv, '
        "2017. https://doi.org/10.48550/arXiv.1706.03762.",
        "output_2": 'Vaswani, Ashish et al. "Attention Is All You Need." arXiv, 2017, '
        "https://doi.org/10.48550/arXiv.1706.03762. MLA.",
        "preference": 1,
        "annotator_index": 4,
        "dataset": "selfinstruct",
        "datasplit": "eval",
        "time_per_example": 36.8,
        "price_per_example": 0.3,
        "index": 3,
        "n_annotated": 4,
    },
]



import ast
import getpass
import os
from functools import partial
from pathlib import Path

import datasets

CURRENT_DIR = Path(__file__).parent
BASE_DIR = Path(__file__).parents[2]

### API specific ###
API_MAX_CONCURRENCY = int(os.environ.get("API_MAX_CONCURRENCY", 5))

OPENAI_MAX_CONCURRENCY = int(os.environ.get("OPENAI_MAX_CONCURRENCY", 5))
OPENAI_CLIENT_CONFIG_PATH = os.environ.get("OPENAI_CLIENT_CONFIG_PATH", BASE_DIR / "client_configs/openai_configs.yaml")
# the following is for backward compatibility, the recommended way is to use OPENAI_CLIENT_CONFIG_PATH
OPENAI_API_KEYS = os.environ.get("OPENAI_API_KEYS", os.environ.get("OPENAI_API_KEY", None))
if isinstance(OPENAI_API_KEYS, str):
    OPENAI_API_KEYS = OPENAI_API_KEYS.split(",")
OPENAI_ORGANIZATION_IDS = os.environ.get("OPENAI_ORGANIZATION_IDS", None)
if isinstance(OPENAI_ORGANIZATION_IDS, str):
    OPENAI_ORGANIZATION_IDS = OPENAI_ORGANIZATION_IDS.split(",")
#

ANTHROPIC_API_KEY = os.environ.get("ANTHROPIC_API_KEY", None)
ANTHROPIC_MAX_CONCURRENCY = int(os.environ.get("ANTHROPIC_MAX_CONCURRENCY", API_MAX_CONCURRENCY))

GOOGLE_API_KEY = os.environ.get("GOOGLE_API_KEY", None)

COHERE_API_KEY = os.environ.get("COHERE_API_KEY", None)

DATASETS_TOKEN = os.environ.get("DATASETS_TOKEN", None)
HUGGINGFACEHUB_API_TOKEN = os.environ.get("HUGGINGFACEHUB_API_TOKEN", None)
DATASETS_FORCE_DOWNLOAD = os.environ.get("DATASETS_FORCE_DOWNLOAD", False)
########################

IS_ALPACA_EVAL_2 = ast.literal_eval(os.environ.get("IS_ALPACA_EVAL_2", "True"))
ANNOTATOR_CONFIG_AE1 = "alpaca_eval_gpt4"
ANNOTATOR_CONFIG_AE2 = "weighted_alpaca_eval_gpt4_turbo"
DEFAULT_ANNOTATOR_CONFIG = ANNOTATOR_CONFIG_AE2 if IS_ALPACA_EVAL_2 else ANNOTATOR_CONFIG_AE1
DEFAULT_CACHE_DIR = None
EVALUATORS_CONFIG_DIR = CURRENT_DIR / "evaluators_configs"
MODELS_CONFIG_DIR = CURRENT_DIR / "models_configs"


MINIMAL_EVALUATORS = (
    ANNOTATOR_CONFIG_AE2,
    ANNOTATOR_CONFIG_AE1,
    "aviary_gpt4",
    "gpt4",
    "claude",
    "text_davinci_003",
    "chatgpt",
    "lmsys_gpt4",
    "humans",
    "alpaca_farm_greedy_gpt4",
)

VERIFIED_EVALUATORS = tuple(
    list(MINIMAL_EVALUATORS)
    + [
        "claude_ranking",
        "improved_aviary_gpt4",
        "improved_lmsys_gpt4",
        "lmsys_gpt4",
        "cohere",
        "alpaca_farm",
        "alpaca_farm_greedy_gpt4",
        "guanaco_33b",
        "longest",
    ]
)

# order matters i => i+1 when filtering
ORDERED_LEADERBOARD_MODES = ["minimal", "verified", "community"]


def get_alpaca_eval_data(dataset="alpaca_eval_gpt4_baseline"):
    dataset = datasets.load_dataset(
        "tatsu-lab/alpaca_eval",
        dataset,
        cache_dir=DEFAULT_CACHE_DIR,
        token=DATASETS_TOKEN,
        download_mode="force_redownload" if DATASETS_FORCE_DOWNLOAD else None,
    )["eval"]
    return dataset


ALPACAEVAL_REFERENCE_OUTPUTS_2 = get_alpaca_eval_data
ALPACAEVAL_REFERENCE_OUTPUTS_1 = partial(get_alpaca_eval_data, dataset="alpaca_eval")

ALPACAEVAL_REFERENCE_OUTPUTS = ALPACAEVAL_REFERENCE_OUTPUTS_2 if IS_ALPACA_EVAL_2 else ALPACAEVAL_REFERENCE_OUTPUTS_1


def ALPACAFARM_GOLD_CROSSANNOTATIONS():
    df = datasets.load_dataset(
        "tatsu-lab/alpaca_eval",
        "alpaca_farm_human_crossannotations",
        cache_dir=DEFAULT_CACHE_DIR,
        token=DATASETS_TOKEN,
        download_mode="force_redownload" if DATASETS_FORCE_DOWNLOAD else None,
    )["validation"].to_pandas()

    # turkers took around 9 min for 15 examples in AlpacaFarm
    df["time_per_example"] = 9.2 * 60 / 15
    df["price_per_example"] = 0.3  # price we paid for each example
    return df


def ALPACAFARM_GOLD_ANNOTATIONS():
    df = datasets.load_dataset(
        "tatsu-lab/alpaca_eval",
        "alpaca_farm_human_annotations",
        cache_dir=DEFAULT_CACHE_DIR,
        token=DATASETS_TOKEN,
        download_mode="force_redownload" if DATASETS_FORCE_DOWNLOAD else None,
    )["validation"].to_pandas()

    # turkers took around 9 min for 15 examples in AlpacaFarm
    df["time_per_example"] = 9.2 * 60 / 15
    df["price_per_example"] = 0.3  # price we paid for each example
    return df


ALPACAEVAL_2_LEADERBOARD_PATHS = CURRENT_DIR / f"leaderboards/data_AlpacaEval_2"
ALPACAEVAL_1_LEADERBOARD_PATHS = CURRENT_DIR / f"leaderboards/data_AlpacaEval"
ALPACAEVAL_LEADERBOARD_PATHS = ALPACAEVAL_2_LEADERBOARD_PATHS if IS_ALPACA_EVAL_2 else ALPACAEVAL_1_LEADERBOARD_PATHS


PRECOMPUTED_LEADERBOARDS = {
    (str(ALPACAEVAL_REFERENCE_OUTPUTS_1), "claude"): ALPACAEVAL_1_LEADERBOARD_PATHS / "claude_leaderboard.csv",
    (str(ALPACAEVAL_REFERENCE_OUTPUTS_1), ANNOTATOR_CONFIG_AE1): ALPACAEVAL_1_LEADERBOARD_PATHS
    / f"{ANNOTATOR_CONFIG_AE1}_leaderboard.csv",
    (str(ALPACAEVAL_REFERENCE_OUTPUTS_1), "chatgpt_fn"): ALPACAEVAL_1_LEADERBOARD_PATHS / "chatgpt_fn_leaderboard.csv",
    (str(ALPACAEVAL_REFERENCE_OUTPUTS_2), ANNOTATOR_CONFIG_AE2): ALPACAEVAL_2_LEADERBOARD_PATHS
    / f"{ANNOTATOR_CONFIG_AE2}_leaderboard.csv",
    (str(ALPACAEVAL_REFERENCE_OUTPUTS_2), "weighted_alpaca_eval_gpt4_turbo"): ALPACAEVAL_2_LEADERBOARD_PATHS
    / f"weighted_alpaca_eval_gpt4_turbo_leaderboard.csv",
    # (str(ALPACAEVAL_REFERENCE_OUTPUTS_2), "alpaca_eval_cot_gpt4_turbo_fn"): ALPACAEVAL_2_LEADERBOARD_PATHS
    # / f"alpaca_eval_cot_gpt4_turbo_fn_leaderboard.csv",
}

HUMAN_ANNOTATED_MODELS_TO_KEEP = (
    "GPT-4 300 characters",
    "GPT-4",
    "AlpacaFarm PPO sim (step 40)",
    "ChatGPT",
    "ChatGPT 300 characters",
    "AlpacaFarm best-of-16 human",
    "AlpacaFarm PPO sim (gpt4 greedy, step 30)",
    "Davinci003",
    "AlpacaFarm ExpIter human (n=128)",
    "AlpacaFarm SFT 10K",
    "AlpacaFarm PPO human (10k, step 40)",
    "Alpaca 7B",
    "AlpacaFarm FeedMe human",
    "Davinci001",
    "LLaMA 7B",
)

EVALUATORS_LEADERBOARD_COLS_TO_PRIORITIZE = [
    "Human agreement",
    "Price [$/1000 examples]",
    "Time [seconds/1000 examples]",
    "Spearman corr.",
    "Pearson corr.",
    "Bias",
    "Variance",
    "Proba. prefer longer",
    "Proba. prefer lists",
    "Proba. prefer 1",
]

MINIMAL_MODELS_FOR_NEW_LEADERBOARD = [
    "gpt4_turbo",
    "gpt4",
    "tulu-2-dpo-70b",
    "Yi-34B-Chat",
    "llama-2-70b-chat-hf",
    "claude-2.1",
    "cohere",
    "chatgpt",
    "gemini-pro",
    "Mixtral-8x7B-Instruct-v0.1",
    "Mistral-7B-Instruct-v0.2",
    "vicuna-33b-v1.3",
    "alpaca-7b",
]

# maps models to Arena Elo rating
CHATBOT_ARENA_LEADERBOARD = {
    "gpt4_turbo": 1243,
    "gpt4": 1192,
    "tulu-2-dpo-70b": 1110,
    "Yi-34B-Chat": 1110,
    "llama-2-70b-chat-hf": 1077,
    "claude-2.1": 1117,
    "chatgpt": 1074,
    "gemini-pro": 1111,
    "Mixtral-8x7B-Instruct-v0.1": 1121,
    "Mistral-7B-Instruct-v0.2": 1023,
    "vicuna-33b-v1.3": 1095,
}

EVALUATORS_LEADERBOARD_COLS_TO_PRINT = EVALUATORS_LEADERBOARD_COLS_TO_PRIORITIZE[:8]

CURRENT_USER = getpass.getuser()
if CURRENT_USER in ["yanndubs"]:
    DEFAULT_CACHE_DIR = "/juice5/scr5/nlp/crfm/human-feedback/cache"


def ALPACAFARM_ALL_OUTPUTS():
    if IS_ALPACA_EVAL_2:
        return [f"results/{m}/model_outputs.json" for m in MINIMAL_MODELS_FOR_NEW_LEADERBOARD]
    else:
        return datasets.load_dataset(
            "tatsu-lab/alpaca_eval",
            "alpaca_eval_all_outputs",
            cache_dir=DEFAULT_CACHE_DIR,
            token=DATASETS_TOKEN,
            download_mode="force_redownload" if DATASETS_FORCE_DOWNLOAD else None,
        )["eval"]



"""
Helper classes for processing the data. Each of those should have a function preprocess and postprocess, which will
respectively be called in SingleAnnotator._preprocess and SingleAnnotator._postprocess in reverse order.

Note: not worth to make the changes but all the parsers could have been processors.
"""

import abc
from typing import Optional, Sequence

import numpy as np
import pandas as pd

from . import utils

__all__ = ["RandomSwitchTwoColumnsProcessor", "PaddingForBatchesProcessor"]


class BaseProcessor(abc.ABC):
    """Base class for a processor."""

    def __init__(self, seed: int = 123):
        self.seed = seed

    @abc.abstractmethod
    def preprocess(self, df_to_annotate: pd.DataFrame) -> pd.DataFrame:
        """Process the annotation dataframe before annotations."""
        pass

    @abc.abstractmethod
    def postprocess(self, df_annotated: pd.DataFrame) -> pd.DataFrame:
        """Process the annotation dataframe after annotations."""
        pass


class RandomSwitchTwoColumnsProcessor(BaseProcessor):
    r"""Randomly switch the order of two columns.

    Parameters
    ----------
    two_columns_to_switch : Sequence[str]
        The two columns to switch.

    fn_replace_if_switch : Optional[Callable[[pd.DataFrame], pd.DataFrame]], optional
        Function to apply to the dataframe formed of the rows with a switch. By default, does nothing.

    fn_replace_if_unswitch : Optional[Callable[[pd.DataFrame], pd.DataFrame]], optional
        Function to apply to the dataframe formed of the rows without a switch. By default, applies the same as
        `fn_replace_if_switch`.

    random_seed_columns : Optional[Sequence[str]], optional
        The columns to use to seed the random choice of switching or not. If None, will use `columns_to_switch`.

    kwargs :
        Additional arguments to pass to `BaseProcessor`. E.g. seed.

    Examples
    --------
    >>> df = pd.DataFrame([dict(instruction='2+2', output_1='10', output_2='4', preference=2),
    ...                    dict(instruction='2+3', output_1='5', output_2='7', preference=1)])
    >>> processor = RandomSwitchTwoColumnsProcessor(two_columns_to_switch=['output_1', 'output_2'],
    ...                                             fn_replace_if_switch = lambda x: x.replace({"preference":{1: 2, 2: 1}}))
    >>> processor.preprocess(df)
        instruction output_1 output_2  preference is_switch_output_1_output_2
    0         2+2         4       10           1                         True
    1         2+3         5        7           1                        False
    >>> (processor.postprocess(processor.preprocess(df)) == df).all(axis=None)
    True
    """

    def __init__(
        self,
        two_columns_to_switch: Sequence[str],
        fn_replace_if_switch=None,
        fn_replace_if_unswitch=None,
        random_seed_columns: Optional[Sequence[str]] = None,
        _switch_column: Optional[str] = None,
        **kwargs,
    ):
        self.two_columns_to_switch = list(set(two_columns_to_switch))
        if len(self.two_columns_to_switch) != 2:
            raise ValueError(
                f"two_columns_to_switch should have exactly two different columns but {two_columns_to_switch}"
            )
        self.fn_replace_if_switch = fn_replace_if_switch or (lambda x: x)
        # by default we assume that it's an involutive function
        self.fn_replace_if_unswitch = fn_replace_if_unswitch or self.fn_replace_if_switch

        # `switch_column` used for backward compatibility
        if _switch_column is None:
            _switch_column = "_".join(["is_switch"] + list(two_columns_to_switch))
        self._switch_column = _switch_column

        if random_seed_columns is None:
            random_seed_columns = two_columns_to_switch
        self.random_seed_columns = sorted(list(random_seed_columns))

        super().__init__(**kwargs)

    def preprocess(self, df_to_annotate: pd.DataFrame) -> pd.DataFrame:
        """When preprocessing, we select the rows to switch and perform the switch."""
        df_to_annotate = df_to_annotate.copy()

        # randomize order of output_1, output_2 base on inputs
        df_to_annotate[self._switch_column] = df_to_annotate.apply(
            # we add "_switch_column" at the beginning to not use the same seed for all tasks
            lambda x: utils.random_seeded_choice(
                seed=self._switch_column + "".join(x[self.random_seed_columns]) + str(self.seed),
                choices=[False, True],
            ),
            axis=1,
        )
        return self._switch_or_unswitch(df_to_annotate, is_switch=True)

    def postprocess(self, df_annotated: pd.DataFrame) -> pd.DataFrame:
        """When postprocessing, we undo the switch and remove the switch column."""
        df_annotated = df_annotated.copy()
        df_annotated = self._switch_or_unswitch(df_annotated, is_switch=False)
        df_annotated = df_annotated.drop(columns=[self._switch_column])
        return df_annotated

    @property
    def col1(self):
        return self.two_columns_to_switch[0]

    @property
    def col2(self):
        return self.two_columns_to_switch[1]

    def _switch_or_unswitch(self, df: pd.DataFrame, is_switch: bool) -> pd.DataFrame:
        """Applies the switch to the dataframe. If `is_switch=False` will undo the switch."""

        # switching two columns is an involution => no need to use is_switch here
        col1_values = df[self.col1].copy()
        col2_values = df[self.col2].copy()
        is_switch_arr = df[self._switch_column]
        df[self.col2] = np.where(is_switch_arr, col1_values, col2_values)
        df[self.col1] = np.where(is_switch_arr, col2_values, col1_values)

        if is_switch:
            df.loc[is_switch_arr, :] = self.fn_replace_if_switch(df.loc[is_switch_arr, :])
        else:
            df.loc[is_switch_arr, :] = self.fn_replace_if_unswitch(df.loc[is_switch_arr, :])

        return df


class PaddingForBatchesProcessor(BaseProcessor):
    r"""Pad the dataframe to have a number of examples divisible by `batch_size`.

    Parameters
    ----------
    batch_size : int
        Number of examples to batch in a single prompt.

    padding_example : dict
        Padding example to use if len(df) not divisible by batch_size.

    kwargs :
        Additional arguments to pass to `BaseProcessor`. E.g. seed.

    Examples
    --------
    >>> df = pd.DataFrame({"instruction": ["solve", "write", "other 1"],
    ...                    "input": ["1+1", "'abc'", ""]})
    >>> processor = PaddingForBatchesProcessor(batch_size=2, padding_example=dict(instruction="pad", input="pad_in"))
    >>> processor.preprocess(df)
        instruction   input  is_padding
    0         solve     1+1       False
    1         write   'abc'       False
    2       other 1               False
    3           pad  pad_in        True
    >>> (processor.postprocess(processor.preprocess(df)) == df).all(axis=None)
    True
    """

    def __init__(self, batch_size, padding_example: dict, **kwargs):
        self.batch_size = batch_size
        self.padding_example = padding_example
        super().__init__(**kwargs)

    def preprocess(self, df_to_annotate: pd.DataFrame) -> pd.DataFrame:
        # padding if you don't have enough examples
        n_to_pad = (self.batch_size - len(df_to_annotate)) % self.batch_size
        padding = pd.DataFrame([self.padding_example] * n_to_pad)
        padding["is_padding"] = True
        df_out = pd.concat([df_to_annotate, padding], axis=0, ignore_index=True)
        df_out["is_padding"] = df_out["is_padding"].fillna(False)
        return df_out

    def postprocess(self, df_annotated: pd.DataFrame) -> pd.DataFrame:
        return df_annotated[~df_annotated["is_padding"].astype(bool)].drop(columns=["is_padding"]).copy()



import functools
import logging
import multiprocessing
import time
from typing import Sequence

import numpy as np
import tqdm
from huggingface_hub.inference_api import InferenceApi

from .. import constants, utils

__all__ = ["huggingface_api_completions"]


def huggingface_api_completions(
    prompts: Sequence[str],
    model_name: str,
    gpu: bool = False,
    do_sample: bool = False,
    num_procs: int = 1,
    **kwargs,
) -> dict[str, list]:
    """Decode with the API from hugging face hub.

    Parameters
    ----------
    prompts : list of str
        Prompts to get completions for.

    model_name : str, optional
        Name of the model (repo on hugging face hub)  to use for decoding.

    gpu : bool, optional
        Whether to use GPU for decoding.

    do_sample : bool, optional
        Whether to use sampling for decoding.

    num_procs : int, optional
        Number of parallel processes to use for decoding.

    kwargs :
        Additional kwargs to pass to `InferenceApi.__call__`.
    """
    n_examples = len(prompts)
    if n_examples == 0:
        logging.info("No samples to annotate.")
        return []
    else:
        logging.info(f"Using `huggingface_api_completions` on {n_examples} prompts using {model_name}.")

    inference = InferenceApi(
        model_name,
        task="text-generation",
        token=constants.HUGGINGFACEHUB_API_TOKEN,
        gpu=gpu,
    )

    default_kwargs = dict(do_sample=do_sample, options=dict(wait_for_model=True), return_full_text=False)
    default_kwargs.update(kwargs)
    logging.info(f"Kwargs to completion: {default_kwargs}")

    with utils.Timer() as t:
        partial_completion_helper = functools.partial(inference_helper, inference=inference, params=default_kwargs)
        if num_procs == 1:
            completions = [partial_completion_helper(prompt) for prompt in tqdm.tqdm(prompts, desc="prompts")]
        else:
            with multiprocessing.Pool(num_procs) as p:
                completions = list(
                    tqdm.tqdm(
                        p.imap(partial_completion_helper, prompts),
                        desc="prompts",
                        total=len(prompts),
                    )
                )
    logging.info(f"Time for {n_examples} completions: {t}")

    completions = [completion["generated_text"] for completion in completions]

    # unclear pricing
    price = [np.nan] * len(completions)
    avg_time = [t.duration / n_examples] * len(completions)

    return dict(completions=completions, price_per_example=price, time_per_example=avg_time)


def inference_helper(prompt: str, inference, params, n_retries=100, waiting_time=2) -> dict:
    for _ in range(n_retries):
        output = inference(inputs=prompt, params=params)
        if "error" in output and n_retries > 0:
            error = output["error"]
            if "Rate limit reached" in output["error"]:
                logging.warning(f"Rate limit reached... Trying again in {waiting_time} seconds. Full error: {error}")
                time.sleep(waiting_time)
            elif "Input validation error" in error and "max_new_tokens" in error:
                params["max_new_tokens"] = int(params["max_new_tokens"] * 0.8)
                logging.warning(
                    f"`max_new_tokens` too large. Reducing target length to {params['max_new_tokens']}, " f"Retrying..."
                )
                if params["max_new_tokens"] == 0:
                    raise ValueError(f"Error in inference. Full error: {error}")
            else:
                raise ValueError(f"Error in inference. Full error: {error}")
        else:
            return output[0]
    raise ValueError(f"Error in inference. We tried {n_retries} times and failed.")



import re
import shutil
from pathlib import Path
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
import yaml

from alpaca_eval import constants, main, utils
from alpaca_eval.annotators import SinglePairwiseAnnotator


@pytest.fixture
def model_outputs():
    return [
        {
            "instruction": "1+1",
            "output": "3",
            "dataset": "test",
        },
        {
            "instruction": "2+2",
            "output": "5",
            "dataset": "test",
        },
        {
            "instruction": "2+3",
            "output": "6",
            "dataset": "test",
        },
    ]


@pytest.fixture
def reference_outputs():
    return [
        {
            "instruction": "1+1",
            "output": "2",
            "dataset": "test",
        },
        {
            "instruction": "2+2",
            "output": "4",
            "dataset": "test",
        },
        {
            "instruction": "2+3",
            "output": "5",
            "dataset": "test",
        },
    ]


@pytest.fixture
def expected_annotations():
    return [
        {
            "instruction": "1+1",
            "output_1": "2",
            "dataset": "test",
            "output_2": "3",
            "annotator": "test",
            "preference": 1,
        },
        {
            "instruction": "2+2",
            "output_1": "4",
            "dataset": "test",
            "output_2": "5",
            "annotator": "test",
            "preference": 1,
        },
        {
            "instruction": "2+3",
            "output_1": "5",
            "dataset": "test",
            "output_2": "6",
            "annotator": "test",
            "preference": 1,
        },
    ]


def _get_mock_annotate(preference=None):
    def annotate(df):
        df = df.copy()
        if preference is None:
            df["preference"] = [1] * len(df)
        else:
            df["preference"] = preference
        return df

    mock_function = MagicMock(side_effect=annotate)
    return mock_function


def clean_up():
    # remove the cache
    Path(constants.EVALUATORS_CONFIG_DIR / "test/annotations_seed0_configs").unlink(missing_ok=True)


def test_evaluate_print(model_outputs, reference_outputs, capsys, expected_annotations):
    mock_function = _get_mock_annotate()
    with patch.object(SinglePairwiseAnnotator, "__call__", mock_function):
        main.evaluate(model_outputs, reference_outputs, annotators_config="test", is_avoid_reannotations=False)

        # Capture the stdout
        captured = capsys.readouterr()
        printed_string = captured.out.strip()
        printed_string = re.sub(r"\s+", " ", printed_string)

        assert printed_string == "win_rate standard_error n_total avg_length Current model 0.00 0.00 3 1"

    clean_up()


def test_evaluate_basic(model_outputs, reference_outputs, expected_annotations):
    mock_function = _get_mock_annotate()
    name = "Current model"
    with patch.object(SinglePairwiseAnnotator, "__call__", mock_function):
        df_leaderboard, annotations = main.evaluate(
            model_outputs,
            reference_outputs,
            annotators_config="test",
            is_return_instead_of_print=True,
            is_avoid_reannotations=False,
        )
        assert annotations == expected_annotations
        assert isinstance(df_leaderboard, pd.DataFrame)
        # win rate is 1 because we always give preference to second which is model output
        assert df_leaderboard.loc[name, "win_rate"] == 0
        assert df_leaderboard.loc[name, "standard_error"] == 0
        assert df_leaderboard.loc[name, "n_total"] == 3
        assert df_leaderboard.loc[name, "mode"] == "community"

    clean_up()


def test_evaluate_advanced(model_outputs, reference_outputs, expected_annotations):
    mock_function = _get_mock_annotate()
    current_leaderboard_mode = "verified"
    name = "current"
    with patch.object(SinglePairwiseAnnotator, "__call__", mock_function):
        df_leaderboard, annotations = main.evaluate(
            model_outputs,
            reference_outputs,
            annotators_config="test",
            is_return_instead_of_print=True,
            is_avoid_reannotations=False,
            current_leaderboard_mode=current_leaderboard_mode,
            name=name,
            max_instances=2,
            precomputed_leaderboard=constants.ALPACAEVAL_LEADERBOARD_PATHS / "alpaca_eval_gpt4_leaderboard.csv",
        )
        assert annotations == expected_annotations[:2]
        assert isinstance(df_leaderboard, pd.DataFrame)
        # win rate is 0 because we always give preference to first which is reference
        assert df_leaderboard.loc[name, "win_rate"] == 0
        assert df_leaderboard.loc[name, "standard_error"] == 0
        assert df_leaderboard.loc[name, "n_total"] == 2
        assert df_leaderboard.loc[name, "mode"] == current_leaderboard_mode
        assert len(df_leaderboard) > 5  # entire leaderboard, 5 is arbitrary here

    clean_up()


def test_analyze_evaluators():
    # make a temporary config file called test
    configs = utils.load_configs(constants.EVALUATORS_CONFIG_DIR / "claude")
    evaluator_name = "tmp_test"
    tmp_path = constants.EVALUATORS_CONFIG_DIR / evaluator_name
    try:
        tmp_path.mkdir(parents=True, exist_ok=False)
        with open(tmp_path / "configs.yaml", "w") as f:
            yaml.dump(configs, f)

        # mock by always returning preference 1
        mock_function = _get_mock_annotate()
        with patch.object(SinglePairwiseAnnotator, "__call__", mock_function):
            df_leaderboard, all_crossannotations = main.analyze_evaluators(
                annotators_config=evaluator_name,
                is_return_instead_of_print=True,
                current_leaderboard_mode="community",
                is_single_annotator=True,
                max_instances=100,
            )
            assert evaluator_name in df_leaderboard.index
            assert df_leaderboard.loc[evaluator_name, "mode"] == "community"
            assert df_leaderboard.loc[evaluator_name, "Human agreement"] < 60
            assert isinstance(all_crossannotations, dict)
            assert isinstance(all_crossannotations[evaluator_name], pd.DataFrame)
            assert len(all_crossannotations[evaluator_name]) == 100
    finally:
        shutil.rmtree(tmp_path)



import functools
import logging
import multiprocessing
from typing import Sequence

import numpy as np
import replicate
import tqdm

from .. import utils

__all__ = ["replicate_completions"]


def replicate_completions(
    prompts: Sequence[str],
    model_name: str,
    num_procs: int = 32,
    **decoding_kwargs,
) -> dict[str, list]:
    r"""Get completions using a model hosted on https://replicate.com/.

    Parameters
    ----------
    prompts : list of str
        Prompts to get completions for.

    model_name : str
        Name of the model endpoint on replicate. Format: owner/name:version

    num_procs : int, optional
        Number of parallel processes to use for decoding.

    decoding_kwargs :
        Additional kwargs to pass to `replicate.run(input={...})`. E.g. `temperature`, `max_length`, `top_p`, etc.
    """
    n_examples = len(prompts)
    if n_examples == 0:
        logging.info("No samples to annotate.")
        return []
    else:
        logging.info(f"Using `replicate` on {n_examples} prompts using {model_name} using {num_procs} processes.")

    logging.info(f"Kwargs to completion: {decoding_kwargs}")

    with utils.Timer() as t:
        partial_completion_helper = functools.partial(
            _replicate_completion_helper, model_name=model_name, **decoding_kwargs
        )
        if num_procs == 1:
            completions = [
                partial_completion_helper(model_name, prompt, **decoding_kwargs)
                for prompt in tqdm.tqdm(prompts, desc="prompts")
            ]
        else:
            with multiprocessing.Pool(num_procs) as p:
                completions = list(
                    tqdm.tqdm(
                        p.imap(partial_completion_helper, prompts),
                        desc="prompts",
                        total=len(prompts),
                    )
                )
    logging.info(f"Time for {n_examples} completions: {t}")

    # unclear pricing because it will depend on
    price = [np.nan] * len(completions)
    avg_time = [t.duration / n_examples] * len(completions)

    return dict(completions=completions, price_per_example=price, time_per_example=avg_time)


def _replicate_completion_helper(prompt: str, model_name: str, **decoding_kwargs):
    """Get a single generation."""
    return "".join(
        replicate.run(
            model_name,
            input={"prompt": prompt, **decoding_kwargs},
        )
    ).strip()



import functools
import logging
import multiprocessing
import random
import time
from typing import Optional, Sequence, Union

import google.generativeai as genai
import numpy as np
import tqdm

from .. import constants, utils

__all__ = ["google_completions"]


def google_completions(
    prompts: Sequence[str],
    max_output_tokens: Union[int, Sequence[int]] = 2048,
    model_name="gemini-pro",
    num_procs: int = constants.API_MAX_CONCURRENCY,
    **decoding_kwargs,  # ,
) -> dict[str, list]:
    """Decode with Anthropic API.

    Parameters
    ----------
    prompts : list of str
        Prompts to get completions for.

    max_output_tokens : int or list of int, optional
        Number of tokens to sample for each prompt. If a list, must be the same length as `prompts`.

    model_name : str, optional
        Name of the model to use for decoding.

    num_procs : int, optional
        Number of parallel processes to use for decoding.

    decoding_kwargs :
        Additional kwargs to pass to `genai.types.GenerationConfig`.
    """
    num_procs = num_procs or constants.API_MAX_CONCURRENCY

    n_examples = len(prompts)
    if n_examples == 0:
        logging.info("No samples to annotate.")
        return dict(completions=[], price_per_example=[], time_per_example=[], completions_all=[])
    else:
        to_log = f"Using `google_completions` on {n_examples} prompts using {model_name} and num_procs={num_procs}."
        logging.info(to_log)

    if isinstance(max_output_tokens, int):
        max_output_tokens = [max_output_tokens] * n_examples

    inputs = zip(prompts, max_output_tokens)

    kwargs = dict(model_name=model_name, **decoding_kwargs)
    kwargs_to_log = {k: v for k, v in kwargs.items() if "api_key" not in k}
    logging.info(f"Kwargs to completion: {kwargs_to_log}")
    with utils.Timer() as t:
        if num_procs == 1:
            responses = [_google_completion_helper(inp, **kwargs) for inp in tqdm.tqdm(inputs, desc="prompts")]
        else:
            with multiprocessing.Pool(num_procs) as p:
                partial_completion_helper = functools.partial(_google_completion_helper, **kwargs)
                responses = list(
                    tqdm.tqdm(
                        p.imap(partial_completion_helper, inputs),
                        desc="prompts",
                        total=len(prompts),
                    )
                )
    logging.info(f"Completed {n_examples} examples in {t}.")

    # anthropic doesn't return total tokens but 1 token approx 4 chars
    price = [_get_price(len(p), len(r), model_name) for p, r in zip(prompts, responses)]

    avg_time = [t.duration / n_examples] * len(responses)

    return dict(completions=responses, price_per_example=price, time_per_example=avg_time, completions_all=responses)


def _google_completion_helper(
    args: tuple[str, int],
    sleep_time: int = 2,
    temperature: Optional[float] = 0.7,
    model_name: str = "gemini-pro",
    google_api_keys: Optional[Sequence[str]] = None,
    max_tries=10,
    **kwargs,
):
    prompt, max_output_tokens = args

    google_api_keys = google_api_keys or (constants.GOOGLE_API_KEY,)
    google_api_key = random.choice(google_api_keys)

    genai.configure(api_key=google_api_key)
    model = genai.GenerativeModel(model_name)
    n_tries = 0

    while True:
        try:
            response = model.generate_content(
                prompt,
                generation_config=genai.types.GenerationConfig(
                    temperature=temperature,
                    max_output_tokens=max_output_tokens,
                    **kwargs,
                ),
                # don't block anything for evaluation
                safety_settings={
                    "HARM_CATEGORY_HARASSMENT": "block_none",
                    "HARM_CATEGORY_DANGEROUS_CONTENT": "block_none",
                    "HARM_CATEGORY_HATE_SPEECH": "block_none",
                    "HARM_CATEGORY_SEXUALLY_EXPLICIT": "block_none",
                },
            )
            text = response.text
            # num_tokens = model.count_tokens(text)

            return text

        # error code 429 is rate limit
        except Exception as e:
            if "429" in str(e):
                logging.info(f"Rate limit reached. Sleeping {sleep_time} seconds.")
                time.sleep(sleep_time)

            else:
                # TODO: better catching of errors when rate limits
                logging.exception(f"Unknown error, so we are retrying. Retry #{n_tries}/{max_tries}. Error:")
                time.sleep(sleep_time)
                n_tries += 1
                if n_tries > max_tries:
                    break

    return ""


def _get_price(n_in_char: int, n_out_char: int, model: str) -> float:
    """Returns the price per token for a given model"""
    if model == "gemini-pro":
        return (n_in_char * 0.00025 + n_out_char * 0.0005) / 1000

    else:
        logging.warning(f"Unknown model {model} for computing price per token.")
        return np.nan



import json
import logging
import multiprocessing
import os
import time
from functools import partial
from typing import Optional, Sequence

import requests

from .. import utils
from .openai import _prompt_to_chatml

__all__ = ["jina_chat_completions"]


def jina_chat_completions(
    prompts: Sequence[str],
    num_procs: Optional[int] = 4,
) -> dict[str, list]:
    """Get jina chat completions for the given prompts. Allows additional parameters such as tokens to avoid and
    tokens to favor.

    Parameters
    ----------
    prompts : list of str
        Prompts to get completions for.
    num_procs : int, optional
        Number of parallel processes to use for decoding.
    """
    n_examples = len(prompts)
    api_key = os.environ.get("JINA_CHAT_API_KEY")

    if n_examples == 0:
        logging.info("No samples to annotate.")
        return {}
    else:
        logging.info(f"Using `jina_chat_completions` on {n_examples} prompts.")

    prompts = [_prompt_to_chatml(prompt.strip()) for prompt in prompts]
    num_processes = min(multiprocessing.cpu_count(), num_procs)
    with utils.Timer() as t:
        with multiprocessing.Pool(processes=num_processes) as pool:
            logging.info(f"Number of processes: {pool._processes}")
            get_chat_completion_with_key = partial(_get_chat_completion, api_key)
            completions_and_num_tokens = pool.map(get_chat_completion_with_key, prompts)

    completions = [text for text, _ in completions_and_num_tokens]
    num_tokens = [tokens for _, tokens in completions_and_num_tokens]

    logging.info(f"Completed {n_examples} examples in {t}.")

    # refer to https://chat.jina.ai/billing
    price_per_example = [0.08 if msg_tokens > 300 else 0 for msg_tokens in num_tokens]
    avg_time = [t.duration / n_examples] * len(completions)

    return dict(completions=completions, price_per_example=price_per_example, time_per_example=avg_time)


def _get_chat_completion(api_key, prompt):
    url = "https://api.chat.jina.ai/v1/chat/completions"
    headers = {"authorization": f"Bearer {api_key}", "content-type": "application/json"}
    json_payload = {"messages": prompt}

    max_retries = 10

    for attempt in range(max_retries):
        try:
            response = requests.post(url, headers=headers, json=json_payload)
            response.raise_for_status()  # Will raise an HTTPError if one occurred.
            message = response.json()["choices"][0]["message"]["content"]
            message_tokens = response.json()["usage"]["completion_tokens"]
            return message, message_tokens
        except (json.JSONDecodeError, requests.exceptions.HTTPError) as e:
            logging.warning(f"Error occurred: {e}, Attempt {attempt + 1} of {max_retries}")
            time.sleep(5)
            if attempt + 1 == max_retries:
                logging.exception("Max retries reached. Raising exception.")
                logging.exception(f"Request data -> URL: {url}, Headers: {headers}, JSON Payload: {json_payload}")
                raise
        except Exception as e:
            logging.exception(f"An unexpected error occurred: {e}")
            raise



import os
import pathlib
from typing import Any, Sequence, Union

import datasets
import pandas as pd

# don't load from utils to avoid unnecessary dependencies
AnyPath = Union[str, os.PathLike, pathlib.Path]
AnyData = Union[Sequence[dict[str, Any]], pd.DataFrame, datasets.Dataset]



import contextlib
import copy
import glob
import itertools
import logging
import os
import pathlib
import random
import re
import sys
import time
from collections import Counter
from pathlib import Path
from typing import Any, Callable, Optional, Sequence, Union

import datasets
import numpy as np
import numpy.typing as npt
import pandas as pd
import pkg_resources
import tqdm
import yaml

from . import constants

# don't load from utils to avoid unnecessary dependencies
AnyPath = Union[str, os.PathLike, pathlib.Path]
AnyData = Union[Sequence[dict[str, Any]], pd.DataFrame, datasets.Dataset]
DUMMY_EXAMPLE = dict(instruction="1+1=", output_1="2", input="", output_2="3")


def read_or_return(to_read: Union[AnyPath, str], **kwargs):
    """Read a file or return the input if it is already a string."""
    try:
        with open(Path(to_read), **kwargs) as f:
            out = f.read()
    except FileNotFoundError as e:
        if Path(to_read).is_absolute():
            # The path is not absolute, so it's not just a string
            raise e

        logging.warning(f"Returning input because file not found. Error: {e}")
        out = to_read

    return out


def random_seeded_choice(seed: Union[int, str, float], choices, **kwargs):
    """Random choice with a (potentially string) seed."""
    return random.Random(seed).choices(choices, k=1, **kwargs)[0]


def is_derangement(arr1, arr2):
    """Whether 2 arrays are derangements of one another"""
    return all([a != b for a, b in zip(arr1, arr2)])


def random_derangement(arr, max_loop=10, seed=None):
    """
    Make random derangement of an array. I.e. shuffle without keeping any element in place. To be efficient,
    we first try `max_loop` rejection sampling. If didn't work then computes all possible derangement.
    """
    if len(arr) < 2:
        return arr

    rng = random.Random(seed)

    idcs = list(range(len(arr)))
    shuffled = list(range(len(arr)))

    for _ in range(max_loop):
        rng.shuffle(shuffled)
        if is_derangement(idcs, shuffled):
            return arr[shuffled]

    # if no luck then computes all possibilities
    deranged_order = list(set([s for s in itertools.permutations(idcs) if is_derangement(s, idcs)]))
    return arr[list(rng.choice(deranged_order))]


def _find_first_match(text: str, outputs_to_match: dict[str, Any]) -> tuple[Any, Any]:
    """Given text to parse and a dictionary of compiled regex to match, return the first match and corresponding key."""
    first_match = None
    first_key = None

    for key, compiled_regex in outputs_to_match.items():
        match = compiled_regex.search(text)
        if match and (not first_match or match.start() < first_match.start()):
            first_match = match
            first_key = key

    return first_match, first_key


def make_prompts(
    df: pd.DataFrame,
    template: str,
    batch_size: int = 1,
) -> tuple[list[str], pd.DataFrame]:
    r"""Helper function to make batch prompts for a single template.

    Parameters
    ----------
    df : pd.DataFrame
        Examples to annotate

    template : str
        Template for the prompt. Should have batch_size number of placeholder {key} where key is a column in df.

    batch_size : int
        Number of examples to batch in a single prompt.

    Returns
    -------
    prompts : list[str]
        List of formatted prompts.

    df_out : pd.DataFrame
        All examples. Will be df with potential padding examples.

    Example
    -------
    >>> import pandas as pd
    >>> df = pd.DataFrame({"instruction": ["solve", "write backwards", "other 1", "pad"],
    ...                    "input": ["1+1", "'abc'", "", "pad_in"]})
    >>> make_prompts(df, template="first: {instruction} {input}, second: {instruction} {input}", batch_size=2)[0]
    ["first: solve 1+1, second: write backwards 'abc'", 'first: other 1 , second: pad pad_in']
    """

    if df.empty:
        return [], df

    text_to_format = re.findall(r"{([^ \s]+?)}", template)
    n_occurrences = Counter(text_to_format)

    if not all([n == batch_size for n in n_occurrences.values()]):
        raise ValueError(f"All placeholders should be repeated batch_size={batch_size} times but {n_occurrences}.")

    if len(df) % batch_size > 0:
        raise ValueError(
            f"The number of rows should be dividable by the batch_size={batch_size} but got {len(df)}."
            "You should use PaddingForBatchesProcessor"
        )

    df_out = df.copy()
    prompts = []
    # ugly for loops, not trivial to vectorize because of the batching
    for i in range(0, len(df_out), batch_size):
        current_prompt = copy.deepcopy(template)
        for j in range(batch_size):
            for to_format in n_occurrences.keys():
                # replace only first occurrence (that's why we don't use .format)
                current_prompt = current_prompt.replace("{" + to_format + "}", str(df_out.iloc[i + j][to_format]), 1)
        prompts.append(current_prompt)

    return prompts, df_out


def convert_ordinal_to_binary_preference(
    preferences: Union[pd.DataFrame, list[dict[str, Any]]],
    ordinal_preference_key: str = "preference",
    binary_preference_key: str = "preference",
):
    """Convert ordinal preference annotations to preference annotations. By merging multiple subcategories together,
    eg A/a/b/B into A/B, or AA/A/a/b/B/BB into A/B.

    Parameters
    ----------
    preferences : pd.DataFrame or list of dicts
        List of dictionaries or a dataframe that contains ordinal preference A/a/b/B in ordinal_preference_key.

    ordinal_preference_key : str
        Key in the dictionaries or column name of the ordinal preference annotations.

    binary_preference_key : str
        Key in the dictionaries or column name of the binary preference annotations. This can be the same
        as ordinal_preference_key if you want to overwrite the ordinal preference annotations.

    Returns
    -------
    binary_preferences
        List of dictionary or a dataframe (same type as the input) that contains binary preferences A/B in
        binary_preference_key.

    Examples
    --------
    >>> from alpaca_eval.utils import convert_ordinal_to_binary_preference
    >>> preferences = [dict(output="test A", preference=1),
    ...                dict(output="test a", preference=2),
    ...                dict(output="test b", preference=3),
    ...                dict(output="test B", preference=4),
    ...                dict(output="test None", preference=0)]
    >>> convert_ordinal_to_binary_preference(preferences, ordinal_preference_key="preference",
    ...     binary_preference_key="preference")
    [{'output': 'test A', 'preference': 1},
     {'output': 'test a', 'preference': 1},
     {'output': 'test b', 'preference': 2},
     {'output': 'test B', 'preference': 2},
     {'output': 'test None', 'preference': 0}]
    """
    if isinstance(preferences, pd.DataFrame):
        is_df = True
    else:
        is_df = False
        preferences = pd.DataFrame.from_records(preferences)

    preferences[binary_preference_key] = (preferences[ordinal_preference_key].round().astype(int) - 1) // 2 + 1

    if not is_df:
        preferences = preferences.to_dict(orient="records")

    return preferences


def convert_to_dataframe(data: AnyData) -> pd.DataFrame:
    """Convert input that AlpacaEval accepts into a dataframe."""
    if isinstance(data, pd.DataFrame):
        return data.copy()
    elif isinstance(data, datasets.Dataset):
        return data.data.to_pandas()
    elif isinstance(data, list):
        return pd.DataFrame.from_records(data)
    else:
        # try
        return pd.DataFrame(data)


def check_imports(modules: Sequence[str], to_use: str = "this fnction"):
    """Check whether the given module is imported."""
    modules = list(modules)
    for module in modules:
        if module not in sys.modules:
            error = f"You need {modules} to use {to_use}. Try `pip install {' '.join(modules)}`."
            raise ImportError(error)


def check_pkg_atleast_version(package, atleast_version):
    curr_version = pkg_resources.get_distribution(package).version
    return pkg_resources.parse_version(curr_version) > pkg_resources.parse_version(atleast_version)


def load_or_convert_to_dataframe(df=Union[AnyPath, AnyData, Callable, list, tuple], **kwargs):
    """Load a dataframe from a path or convert the input to a dataframe if it's not a path."""
    if isinstance(df, Callable):
        df = df(**kwargs)

    if isinstance(df, (tuple, list)) and isinstance(df[0], (str, os.PathLike, pathlib.Path)):
        df = pd.concat(
            [load_or_convert_to_dataframe(f, **kwargs) for f in df],
        )

    if isinstance(df, (str, os.PathLike, pathlib.Path)):
        df = Path(df)

        # check if it's a globbing pattern
        if "*" in str(df):
            df = pd.concat(
                [load_or_convert_to_dataframe(f, **kwargs) for f in glob.glob(str(df))],
            )
        else:
            suffix = df.suffix
            if suffix == ".json":
                df = pd.read_json(df, **kwargs)
            elif suffix == ".csv":
                df = pd.read_csv(df, **kwargs)
                if df.columns[0] == "Unnamed: 0":
                    df.set_index(df.columns[0], inplace=True)
                    df.index.name = None
            elif suffix == ".tsv":
                df = pd.read_table(df, sep="\t", **kwargs)
            else:
                raise ValueError(f"File format {suffix} not supported.")
    else:
        df = convert_to_dataframe(df, **kwargs)

    return df


class Timer:
    """Timer context manager"""

    def __enter__(self):
        """Start a new timer as a context manager"""
        self.start = time.time()
        return self

    def __exit__(self, *args):
        """Stop the context manager timer"""
        self.end = time.time()
        self.duration = self.end - self.start

    def __str__(self):
        return f"{self.duration:.1f} seconds"


@contextlib.contextmanager
def silent():
    """Context manager to remove all outputs and warnings."""
    import IPython

    with open(os.devnull, "w") as f, contextlib.redirect_stdout(f), DisableLogger(), IPython.utils.io.capture_output():
        yield


class DisableLogger:
    def __enter__(self):
        logging.disable(50)

    def __exit__(self, a, b, c):
        logging.disable(logging.NOTSET)


def contains_list(text):
    """Check if the text contains a list / bullet points...."""

    # Bullet points or '*' list items
    bullet_point_pattern = r"(\s*•\s*|\s*\*\s*)(\w+)"

    # Numbered lists with '.' or ')'
    number_list_pattern = r"(\s*\d+\.|\s*\d+\))\s*(\w+)"

    # Alphabetic lists with '.' or ')'
    alpha_list_pattern = r"(\s*[a-zA-Z]\.|\s*[a-zA-Z]\))\s*(\w+)"

    # List items starting with a dash '-'
    dash_list_pattern = r"(\s*-\s*)(\w+)"

    patterns = [
        bullet_point_pattern,
        number_list_pattern,
        alpha_list_pattern,
        dash_list_pattern,
    ]

    for pattern in patterns:
        if re.search(pattern, text):
            return True

    return False


def prioritize_elements(lst: list, elements: Sequence) -> list:
    """Prioritize elements in a list. If elements are not in the list, they will be appended to the end of the list."""
    elements = list(elements)
    for el in elements:
        if el in lst:
            lst.remove(el)
    return elements + lst


def load_configs(configs: Union[AnyPath, dict], relative_to: Optional[AnyPath] = None):
    """Load the config yaml files, or return if it's already a dict."""
    if not isinstance(configs, dict):
        if relative_to is not None:
            configs = Path(relative_to) / configs
        configs = Path(configs)
        if configs.is_dir():
            configs = configs / "configs.yaml"
        with open(configs, "r") as stream:
            try:
                configs = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                logging.exception(exc)
        assert isinstance(configs, dict)

    return configs


def get_precomputed_leaderboard(precomputed_leaderboard, reference_outputs, annotators_config):
    if precomputed_leaderboard == "auto":
        try:
            precomputed_leaderboard = constants.PRECOMPUTED_LEADERBOARDS[
                (str(reference_outputs), str(annotators_config))
            ]
        except KeyError:
            try:
                if Path(reference_outputs).is_absolute():
                    logging.warning(
                        f"precomputed_leaderboard = 'auto'. But we have found no corresponding leaderboard for"
                        f" {reference_outputs} and {annotators_config}"
                    )
            except:
                logging.warning(f"precomputed_leaderboard = 'auto'. But we have found no corresponding leaderboard")
            precomputed_leaderboard = None

    if precomputed_leaderboard is not None:
        try:
            leaderboard = load_or_convert_to_dataframe(precomputed_leaderboard).to_dict(orient="index")
        except FileNotFoundError:
            logging.warning(f"precomputed_leaderboard = {precomputed_leaderboard} not found => computing from scratch.")
            leaderboard = dict()
    else:
        leaderboard = dict()
    return leaderboard, precomputed_leaderboard


def get_output_path(output_path, model_outputs, name, dflt_dir="results"):
    if output_path == "auto":
        if model_outputs is None:
            output_path = None
        else:
            try:
                if Path(model_outputs).exists():
                    output_path = Path(model_outputs).parent
            except:
                pass

            if output_path == "auto":
                if name is not None:
                    output_path = Path(dflt_dir) / name
                else:
                    output_path = "."
    if output_path is not None:
        output_path = Path(output_path)
        output_path.mkdir(exist_ok=True, parents=True)
    return output_path


def print_leaderboard(df_leaderboard, leaderboard_mode, cols_to_print, current_name=None):
    cols_to_print = list(cols_to_print)

    if leaderboard_mode is not None:
        if "mode" in df_leaderboard.columns:
            # select all modes that come before
            current_idx = constants.ORDERED_LEADERBOARD_MODES.index(leaderboard_mode)
            df_leaderboard["mode_idx"] = df_leaderboard["mode"].apply(constants.ORDERED_LEADERBOARD_MODES.index)

            is_smaller_mode = df_leaderboard["mode_idx"] <= current_idx
            is_selected = is_smaller_mode | (df_leaderboard["mode"].isnull())

            if current_name is not None:
                is_selected |= df_leaderboard.index == current_name

            df_leaderboard = df_leaderboard[is_selected]
    elif "mode" in df_leaderboard.columns:
        cols_to_print = cols_to_print + ["mode"]
    print(df_leaderboard[cols_to_print].to_string(float_format="%.2f"))


def get_generator_name(name, model_outputs):
    if name is None:
        try:
            assert len(model_outputs["generator"].unique()) == 1
            name = model_outputs["generator"].iloc[0]
        except:
            name = "Current model"
    return name


def get_module_attribute(module, func_name):
    """getattr but only if it's in __all__"""
    if func_name in module.__all__:
        return getattr(module, func_name)
    elif hasattr(module, func_name):
        raise AttributeError(
            f"The function {func_name} is not allowed,add it to __all__ if needed."
            f" Available functions: {module.__all__}"
        )
    else:
        raise AttributeError(f"The function {func_name} does not exist. Available functions: {module.__all__}")


def dataframe_chunk_generator(df: pd.DataFrame, chunksize: Optional[int] = None, tqdm_desc: Optional[str] = None):
    """Generator that yields chunks of a dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe to split into chunks.

    chunksize : int, optional
        The size of the chunks. If None, the chunksize will be the length of the dataframe.

    tqdm_desc : bool, optional
        Description to display in the tqdm progress bar. If None, no progress bar will be displayed.
    """
    if chunksize is None:
        chunksize = max(1, len(df))

    iterator = range(0, len(df), chunksize)

    if tqdm_desc is not None:
        iterator = tqdm.tqdm(iterator, desc=tqdm_desc)

    n_iter = len(df) // chunksize

    for i in iterator:
        df_chunk = df.iloc[i : i + chunksize]

        # if many iterations then better to copy the dataframe to avoid memory issues
        if n_iter > 1:
            df_chunk = df_chunk.copy()

        yield df_chunk


def validate_alpacaeval_preference(x: float, is_allow_nan: bool = True) -> bool:
    """Validate the preference annotation."""
    return (1 <= x <= 2) or (is_allow_nan and np.isnan(x))


def get_all_clients(
    client_config_path: AnyPath,
    model_name: str,
    default_client_class: str,
    get_backwards_compatible_configs: Callable,
    backward_compatibility_kwargs: dict = {},
    **kwargs,
) -> list:
    """Returns a list of different kwargs to pass to the client, each element corresponds to one possible client.
    For more information see `client_configs/README.md`.
    """

    client_config_path = Path(client_config_path)
    if client_config_path.is_file():
        with open(client_config_path) as f:
            all_client_configs = yaml.safe_load(f)

        client_configs = []

        if model_name in all_client_configs:
            if "default" in all_client_configs[model_name]:
                assert "default" in all_client_configs, "default client was asked for but not found"
                client_configs = client_configs + all_client_configs["default"]
                # remove "default" from the list of configs for this model
                all_client_configs[model_name] = [
                    config for config in all_client_configs[model_name] if config != "default"
                ]

            client_configs = client_configs + all_client_configs[model_name]

        else:
            assert (
                "default" in all_client_configs
            ), f"default client config is required as there are no model specific configs for {model_name}"
            client_configs = all_client_configs["default"]

    else:
        # backward compatibility
        logging.warning(
            f"{client_config_path} wasn't found. We are using environment variables to construct the client configs."
            "This is the old and non-recommended way of doing it. Please see `client_configs/README.md` for the "
            "recommended way of specifying client configs."
        )
        client_configs = get_backwards_compatible_configs(**backward_compatibility_kwargs)

    all_clients = []
    for config in client_configs:
        client_class = config.pop("client_class", default_client_class)
        ClientClass = import_class(client_class)
        all_clients.append(ClientClass(**config, **kwargs))

    return all_clients


def import_class(full_class_string):
    """
    Dynamically import a class from a string.

    Parameters
    ----------
    full_class_string:
        The full class string. E.g., 'openai.OpenAI' return OpenAI
    """
    module_name, class_name = full_class_string.rsplit(".", 1)

    # Import the module
    module = __import__(module_name, fromlist=[class_name])

    # Get the class
    cls = getattr(module, class_name)

    return cls



import copy
import functools
import json
import logging
import multiprocessing
import time
from typing import Optional, Sequence, Union

import boto3
import botocore.exceptions
import tqdm

from .. import utils

__all__ = ["bedrock_anthropic_completions"]

DEFAULT_NUM_PROCS = 3


def bedrock_anthropic_completions(
    prompts: Sequence[str],
    max_tokens_to_sample: Union[int, Sequence[int]] = 2048,
    model_name: str = "anthropic.claude-v1",
    num_procs: int = DEFAULT_NUM_PROCS,
    **decoding_kwargs,
) -> dict[str, list]:
    """Decode with Anthropic API.

    Parameters
    ----------
    prompts : list of str
        Prompts to get completions for.

    model_name : str, optional
        Name of the model to use for decoding.

    num_procs : int, optional
        Number of parallel processes to use for decoding.

    decoding_kwargs :
        Additional kwargs to pass to Bedrock Anthropic.
    """
    num_procs = num_procs or DEFAULT_NUM_PROCS

    n_examples = len(prompts)
    if n_examples == 0:
        logging.info("No samples to annotate.")
        return []
    else:
        to_log = f"Using `bedrock_anthropic_completions` on {n_examples} prompts using {model_name} and num_procs={num_procs}."
        logging.info(to_log)

    if isinstance(max_tokens_to_sample, int):
        max_tokens_to_sample = [max_tokens_to_sample] * n_examples

    inputs = zip(prompts, max_tokens_to_sample)

    kwargs = dict(model_name=model_name, **decoding_kwargs)
    kwargs_to_log = {k: v for k, v in kwargs.items() if "api_key" not in k}
    logging.info(f"Kwargs to completion: {kwargs_to_log}")
    with utils.Timer() as t:
        if num_procs == 1:
            responses = [
                _bedrock_anthropic_completion_helper(inp, **kwargs) for inp in tqdm.tqdm(inputs, desc="prompts")
            ]
        else:
            with multiprocessing.Pool(num_procs) as p:
                partial_completion_helper = functools.partial(_bedrock_anthropic_completion_helper, **kwargs)
                responses = list(
                    tqdm.tqdm(
                        p.imap(partial_completion_helper, inputs),
                        desc="prompts",
                        total=len(prompts),
                    )
                )
    logging.info(f"Completed {n_examples} examples in {t}.")

    completions = responses

    ## Token counts are not returned by Bedrock for now
    price = [0 for _ in prompts]

    avg_time = [t.duration / n_examples] * len(completions)

    return dict(completions=completions, price_per_example=price, time_per_example=avg_time, completions_all=responses)


def _bedrock_anthropic_completion_helper(
    args: tuple[str, int],
    sleep_time: int = 2,
    region: Optional[str] = "us-west-2",
    model_name: str = "anthropic.claude-v1",
    temperature: Optional[float] = 0.7,
    **kwargs,
):
    prompt, max_tokens = args

    if not utils.check_pkg_atleast_version("boto3", "1.28.58"):
        raise ValueError("boto3 version must be at least 1.28.58 Use `pip install -U boto3`.")

    bedrock = boto3.client(service_name="bedrock-runtime", region_name=region)
    accept = "application/json"
    contentType = "application/json"

    kwargs.update(dict(max_tokens_to_sample=max_tokens, temperature=temperature))
    curr_kwargs = copy.deepcopy(kwargs)
    while True:
        try:
            body = json.dumps({**{"prompt": prompt}, **curr_kwargs})
            response = bedrock.invoke_model(body=body, modelId=model_name, accept=accept, contentType=contentType)
            response = json.loads(response.get("body").read()).get("completion")
            break
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "ThrottlingException":
                logging.warning(f"Hit throttling error: {e}.")
                logging.warning(f"Rate limit hit. Sleeping for {sleep_time} seconds.")
                time.sleep(sleep_time)
        except Exception as e:
            logging.error(f"Hit unknown error : {e}")
            raise e

    return response



import abc
import logging
from dataclasses import dataclass
from numbers import Number
from typing import Sequence, Union

import numpy as np
import numpy.typing as npt
import pandas as pd

from alpaca_eval.utils import validate_alpacaeval_preference


def pairwise_to_winrate(preferences: Union[pd.Series, Sequence]) -> dict[str, float]:
    """Extract head2head metrics (n_wins, n_counts, win_rate) from a sequence preference.
    This assumes that the preference is encoded as 0 or 1.5 for draw, 1 for base win, 2 when the model to compare wins.
    """
    out = AbsoluteScoringRule().describe_head2head(preferences)
    out["discrete_win_rate"] = ZeroOneScoringRule().describe_head2head(preferences)["win_rate"]
    return out


@dataclass
class BaseScoringRule(abc.ABC):
    """Base class for a scoring rule."""

    @abc.abstractmethod
    def _score(self, prediction: pd.Series, target: pd.Series) -> float:
        pass

    @abc.abstractmethod
    def _bayes_estimator(self, predictions: pd.Series) -> Number:
        pass

    def _idcs_draws(self, predictions: pd.Series) -> pd.Series:
        return (predictions == 1.5) | (predictions == 0)

    def score(self, prediction: npt.ArrayLike, target: npt.ArrayLike) -> float:
        """Score a prediction. Higher is better."""
        prediction, target = self.preprocess(prediction, target)
        return self._score(prediction, target)

    def error(self, prediction: npt.ArrayLike, target: npt.ArrayLike) -> float:
        """Compute the error of the prediction. Lower is better"""
        return 1 - self.score(prediction, target)

    def sem(self, prediction: npt.ArrayLike, target: npt.ArrayLike) -> float:
        """Compute the standard error of the error."""
        return pd.Series([self.score([p], [t]) for p, t in zip(prediction, target)]).sem()

    def generalized_win_rate(self, predictions: npt.ArrayLike) -> float:
        """Compute the generalized win rate of the prediction."""
        # predictions = self.preprocess_predictions(predictions)
        # return (predictions - 1).mean() * 100
        return self.describe_head2head(predictions)["win_rate"]

    def describe_head2head(self, predictions: npt.ArrayLike) -> dict[str, float]:
        """Compute the generalized win rate of the prediction."""
        predictions = self.preprocess_predictions(predictions)
        n_draws = self._idcs_draws(predictions).sum()

        # makes it easier to work with
        predictions = predictions.astype(float).replace({0.0: 1.5})

        is_preference = predictions.apply(validate_alpacaeval_preference, is_allow_nan=False)
        n_not_pair = sum(~is_preference)
        if n_not_pair > 0:
            logging.info(f"drop {n_not_pair} outputs that are not preferences")

        predictions = predictions[is_preference] - 1

        n_wins = (predictions > 0.5).sum()
        n_wins_base = (predictions < 0.5).sum()
        n_total = len(predictions)

        return dict(
            win_rate=predictions.mean() * 100,
            standard_error=predictions.sem() * 100,
            n_wins=n_wins,
            n_wins_base=n_wins_base,
            n_draws=n_draws,
            # note that n_draws will happen more often for weighted win rate because you can get 1.5 somewhat often due
            # to float precision
            n_total=n_total,
        )

    def bayes_estimator(self, predictions: npt.ArrayLike) -> Number:
        """Compute the bayes estimator of the predictions."""
        predictions = self.preprocess_predictions(predictions)
        return self._bayes_estimator(predictions)

    def preprocess(self, predictions: npt.ArrayLike, targets: npt.ArrayLike) -> tuple[pd.Series, pd.Series]:
        """Validate the predictions and targets."""
        predictions = self.preprocess_predictions(predictions)
        targets = self.preprocess_targets(targets)

        if predictions.shape != targets.shape:
            raise ValueError(
                f"predictions and targets should have the same shape. Got {predictions.shape} and {targets.shape}"
            )

        return predictions, targets

    def preprocess_predictions(self, predictions: npt.ArrayLike) -> pd.Series:
        """Validate the predictions."""
        # backward compatibility for 0 -> 1.5
        return pd.Series(predictions).replace({0: 1.5}).astype(float)

    def preprocess_targets(self, targets: npt.ArrayLike) -> pd.Series:
        """Validate the targets."""
        # backward compatibility for 0 -> 1.5
        return pd.Series(targets).replace({0: 1.5}).astype(float)


class ZeroOneScoringRule(BaseScoringRule):
    """Scoring rule for binary predictions."""

    def _score(self, prediction, target):
        # accuracy
        return (target == prediction).mean()

    def _bayes_estimator(self, predictions):
        # mode
        return _random_mode(predictions)

    def generalized_win_rate(self, predictions):
        descriptions = self.describe_head2head(predictions)
        assert (
            descriptions["win_rate"]
            == (descriptions["n_wins"] + descriptions["n_draws"] / 2) / descriptions["n_total"] * 100
        ), "generalized_win_rate should be equal to the win_rate for binary predictions"
        return descriptions["win_rate"]

    def preprocess_predictions(self, predictions: npt.ArrayLike) -> pd.Series:
        return pd.Series(predictions).replace({1.5: 0}).round().astype(pd.Int64Dtype())

    def preprocess_targets(self, targets: npt.ArrayLike) -> pd.Series:
        return pd.Series(targets).replace({1.5: 0}).round().astype(pd.Int64Dtype())


class AbsoluteScoringRule(BaseScoringRule):
    """Absolute loss scoring rule (i.e. MAE)."""

    def _score(self, prediction, target):
        # 1 - MAE
        return 1 - (target - prediction).abs().mean()

    def _bayes_estimator(self, predictions):
        # if all the values are 0.0, 1.0, 2.0, nan, or 1.5 then for backward compatibility we return the random mode
        # note that this doesn't change the expected value of the estimator, but increases the variance. The new version
        # is thus better
        if pd.Series(predictions.unique()).isin([0.0, 1.0, 2.0, np.nan, 1.5]).all():
            return _random_mode(predictions)

        return predictions.median()


class SquaredScoringRule(BaseScoringRule):
    """Squared loss scoring rule (i.e. MSE)."""

    def _score(self, prediction, target):
        # 1 - MSE
        return 1 - ((target - prediction) ** 2).mean()

    def _bayes_estimator(self, predictions):
        return predictions.mean()


SCORING_RULES = {
    "zero_one": ZeroOneScoringRule,
    "absolute": AbsoluteScoringRule,
    "squared": SquaredScoringRule,
}


def _random_mode(s, available_modes=None, favorite_mode=None, seed=123, is_dropna=True):
    """Take the mode of a series, but if there are multiple modes, randomly sample one
    (with potential restriction to `available_modes` or favoring a specific mode `favorite_mode`).

    Example
    -------
    >>> import pandas as pd
    >>> from alpaca_eval.metrics import _random_mode
    >>> _random_mode(pd.Series([1.0,2.0,1.0]))
    1.0
    >>> _random_mode(pd.Series([1.0,2.0])) in [1.0, 2.0]
    True
    >>> _random_mode(pd.Series([1.0,2.0,-1.0]), favorite_mode=-1.0)
    -1.0
    >>> _random_mode(pd.Series([1.0,2.0,2.0,-1.0]), favorite_mode=-1.0)
    2.0
    """
    out = pd.Series.mode(s)
    if is_dropna:
        out = out.dropna()

    if len(out) > 1:
        if favorite_mode is not None and favorite_mode in out.values:
            return favorite_mode
        if available_modes is not None:
            out = out[out.isin(available_modes)]
        out = out.sample(1, random_state=seed)

    if len(out) == 0:
        return np.nan

    return out.item()
